#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import decimal

import octobot_commons.logging as logging

import octobot_trading.enums as enums
import octobot_trading.constants as constants
import octobot_trading.personal_data.orders.order as order_class


class TrailingStopOrder(order_class.Order):
    UNINITIALIZED_TRAILING_PERCENT = decimal.Decimal(str(-1))
    DEFAULT_TRAILING_PERCENT = decimal.Decimal(str(5))

    def __init__(self, trader, side=enums.TradeOrderSide.SELL, trailing_percent=UNINITIALIZED_TRAILING_PERCENT):
        super().__init__(trader, side=side)
        self.order_type = enums.TraderOrderType.TRAILING_STOP
        self.trailing_stop_price_hit_event = None
        self.trailing_price_hit_event = None
        self.wait_for_stop_price_hit_event_task = None
        self.wait_for_price_hit_event_task = None
        self.trailing_percent = trailing_percent
        self.allow_instant_fill = True

    async def update_order_status(self, force_refresh=False):
        if not self.trader.simulate and (not self.is_synchronized_with_exchange or force_refresh):
            await self.default_exchange_update_order_status()
        await self._reset_events(self.origin_price, self.creation_time)

    async def set_trailing_percent(self, trailing_percent):
        """
        Set trailing percent and reset event and tasks
        :param trailing_percent: the new trailing percent
        """
        self.trailing_percent = trailing_percent
        await self._reset_events(self.origin_price, self.creation_time)

    async def _reset_events(self, new_price, new_price_time):
        """
        Reset events and tasks
        :param new_price: the new trailing price
        """
        self._clear_event_and_tasks()
        price_events_manager = self.exchange_manager.exchange_symbols_data. \
            get_exchange_symbol_data(self.symbol).price_events_manager
        self._create_hit_events(price_events_manager, new_price, new_price_time)
        await self._handle_events_and_hit_tasks()

    def _create_hit_events(self, price_events_manager, new_price, new_price_time):
        """
        Create prices hit events
        :param price_events_manager: the price events manager to use
        :param new_price: the new trailing price
        """
        if self.trailing_stop_price_hit_event is None:
            self.trailing_stop_price_hit_event = price_events_manager.new_event(
                self._calculate_stop_price(new_price), new_price_time,
                self.side is enums.TradeOrderSide.BUY, self.allow_instant_fill)
        if self.trailing_price_hit_event is None:
            # don't allow instant fill since this event should only be triggered by next recent trades and prices
            self.trailing_price_hit_event = price_events_manager.new_event(new_price, new_price_time,
                                                                           self.side is enums.TradeOrderSide.SELL,
                                                                           allow_instant_fill=False)

    def _calculate_stop_price(self, new_price):
        """
        Calculate the trailing stop price from price and trailing percent
        :param new_price: the price
        :return: the stop price related to the specified price
        """
        trailing_price_factor = (self.trailing_percent
                                 if self.trailing_percent != self.UNINITIALIZED_TRAILING_PERCENT
                                 else self.DEFAULT_TRAILING_PERCENT) / decimal.Decimal("100")
        trailing_price_factor *= constants.ONE if self.side is enums.TradeOrderSide.BUY else -constants.ONE
        return new_price * (constants.ONE + trailing_price_factor)

    async def _handle_events_and_hit_tasks(self):
        """
        Create event hit waiting tasks
        """
        if self.wait_for_price_hit_event_task is None and self.trailing_price_hit_event is not None:
            if self.trailing_price_hit_event.is_set():
                await self._on_price_hit()
            else:
                self.wait_for_price_hit_event_task = asyncio.create_task(
                    _wait_for_price_hit(self.trailing_price_hit_event, self._on_price_hit))

        if self.wait_for_stop_price_hit_event_task is None and self.trailing_stop_price_hit_event is not None:
            if self.trailing_stop_price_hit_event.is_set():
                await self.on_fill()
            else:
                self.wait_for_stop_price_hit_event_task = asyncio.create_task(
                    _wait_for_price_hit(self.trailing_stop_price_hit_event, self.on_fill))

    def _remove_events(self, price_events_manager):
        """
        Remove events from the price events manager
        :param price_events_manager: the price events manager to use
        """
        if self.trailing_stop_price_hit_event is not None:
            price_events_manager.remove_event(self.trailing_stop_price_hit_event)
            self.trailing_stop_price_hit_event = None
        if self.trailing_price_hit_event is not None:
            price_events_manager.remove_event(self.trailing_price_hit_event)
            self.trailing_price_hit_event = None

    def _cancel_hit_tasks(self):
        """
        Cancel and destroy event hit waiting tasks
        """
        if self.wait_for_price_hit_event_task is not None:
            if not self.trailing_price_hit_event.is_set():
                self.wait_for_price_hit_event_task.cancel()
            self.wait_for_price_hit_event_task = None

        if self.wait_for_stop_price_hit_event_task is not None:
            if not self.trailing_stop_price_hit_event.is_set():
                self.wait_for_stop_price_hit_event_task.cancel()
            self.wait_for_stop_price_hit_event_task = None

    async def _on_price_hit(self):
        """
        Is called when the trailing price is hit
        """
        prices_manager = self.exchange_manager.exchange_symbols_data. \
            get_exchange_symbol_data(self.symbol).prices_manager
        logging.get_logger(self.get_logger_name()).debug(f"New price hit {prices_manager.mark_price}, "
                                                         f"replacing stop...")
        await self._reset_events(decimal.Decimal(str(prices_manager.mark_price)), prices_manager.mark_price_set_time)

    async def on_filled(self):
        """
        Create an artificial when trailing stop is filled
        """
        await order_class.Order.on_filled(self)
        # TODO replace with chained order ?
        await self.trader.create_artificial_order(enums.TraderOrderType.SELL_MARKET
                                                  if self.side is enums.TradeOrderSide.SELL
                                                  else enums.TraderOrderType.BUY_MARKET,
                                                  self.symbol, self.origin_stop_price,
                                                  self.origin_quantity, self.origin_stop_price)

    def _clear_event_and_tasks(self):
        """
        Clear prices hit events and their related tasks
        """
        self._cancel_hit_tasks()
        self._remove_events(self.exchange_manager.exchange_symbols_data.
                            get_exchange_symbol_data(self.symbol).price_events_manager)

    def clear(self):
        """
        Clear prices hit events and their related tasks and call super clear
        """
        self._clear_event_and_tasks()
        order_class.Order.clear(self)


async def _wait_for_price_hit(event_to_wait, callback):
    """
    Awaitable wait for event and call the specified callback
    :param event_to_wait: the event to wait
    :param callback: the callback to call
    """
    await asyncio.wait_for(event_to_wait.wait(), timeout=None)
    await callback()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import mock
import trading_backend.exchanges

from tests import event_loop
import octobot_commons.constants as commons_constants
import octobot_commons.configuration as commons_configuration
import octobot_trading.exchanges as exchanges

pytestmark = pytest.mark.asyncio


@pytest.fixture
def tentacles_setup_config():
    setup_config = mock.Mock()
    setup_config.is_tentacle_activated = mock.Mock(return_value=True)
    return setup_config


@pytest.fixture()
def exchange_config():
    return {
        commons_constants.CONFIG_EXCHANGE_KEY: commons_configuration.encrypt("01234").decode(),
        commons_constants.CONFIG_EXCHANGE_SECRET: commons_configuration.encrypt("012345").decode()
    }


async def test_is_compatible_account_with_checked_exchange(exchange_config, tentacles_setup_config):
    with mock.patch.object(trading_backend.exchanges.Binance, "is_valid_account",
                           mock.AsyncMock(return_value=(True, None))) as is_valid_account_mock:
        compatible, error = await exchanges.is_compatible_account("binance", exchange_config, tentacles_setup_config)
        assert compatible is True
        assert error is None
        is_valid_account_mock.assert_called_once()
    with mock.patch.object(trading_backend.exchanges.Binance, "is_valid_account",
                           mock.AsyncMock(return_value=(False, "plop"))) as is_valid_account_mock:
        compatible, error = await exchanges.is_compatible_account("binance", exchange_config, tentacles_setup_config)
        assert compatible is False
        assert "plop" in error and len(error) > len("plop")
        is_valid_account_mock.assert_called_once()


def test_get_partners_explanation_message():
    assert "docs" in exchanges.get_partners_explanation_message()


def test_log_time_sync_error():
    logger = mock.Mock()
    exchanges.log_time_sync_error(logger, "exchange_name", "error", "details")
    args = logger.error.call_args[0][0]
    assert "exchange_name".capitalize() in args
    assert "error" in args
    assert "details" in args
    assert "docs" in args


async def test_is_compatible_account_with_unchecked_exchange(exchange_config, tentacles_setup_config):
    compatible, error = await exchanges.is_compatible_account("hitbtc", exchange_config, tentacles_setup_config)
    assert compatible is False
    assert isinstance(error, str)
    with mock.patch.object(trading_backend.exchanges.Exchange, "is_valid_account",
                           mock.AsyncMock(return_value=(True, "plop"))) as is_valid_account_mock:
        compatible, error = await exchanges.is_compatible_account("hitbtc", exchange_config, tentacles_setup_config)
        assert compatible is True
        assert "plop" in error and len(error) > len("plop")
        is_valid_account_mock.assert_called_once()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

from octobot_commons.errors import ConfigTradingError

# Import required fixtures
from tests import event_loop, install_tentacles
from tests.exchanges import exchange_builder
from octobot_trading.api.exchange import cancel_ccxt_throttle_task

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest.mark.usefixtures("event_loop", "exchange_builder")
async def test_create_without_trading_config(exchange_builder):
    with pytest.raises(ConfigTradingError):
        await exchange_builder.build()


@pytest.mark.usefixtures("event_loop", "exchange_builder")
async def test_create_without_installed_trading_mode(exchange_builder):
    with pytest.raises(ConfigTradingError):
        await exchange_builder.build()


@pytest.mark.usefixtures("event_loop", "exchange_builder", "install_tentacles")
async def test_create_without_installed_trading_mode(exchange_builder):
    with pytest.raises(ConfigTradingError):
        await exchange_builder.build()


@pytest.mark.usefixtures("event_loop", "exchange_builder", "install_tentacles")
async def test_create(exchange_builder):
    # await exchange_builder.build() # TODO
    pass


@pytest.mark.usefixtures("event_loop", "exchange_builder")
@pytest.mark.parametrize("exchange_builder", [(None, "bitmex")], indirect=["exchange_builder"])
async def test_create_basic(exchange_builder):
    exchange_builder.disable_trading_mode()
    exchange_manager = await exchange_builder.build()

    assert exchange_manager is not None
    assert exchange_manager.exchange_name == "bitmex"

    cancel_ccxt_throttle_task()
    await exchange_manager.stop()



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging
import octobot_trading.constants as constants
import octobot_trading.util as util


class FundingManager(util.Initializable):
    def __init__(self):
        super().__init__()
        self.logger = logging.get_logger(self.__class__.__name__)
        self.funding_rate = constants.ZERO
        self.predicted_funding_rate = constants.ZERO
        self.next_update = 0
        self.last_updated = 0
        self.reset_funding()

    async def initialize_impl(self):
        self.reset_funding()

    def reset_funding(self):
        self.funding_rate = constants.ZERO
        self.predicted_funding_rate = constants.ZERO
        self.next_update = 0
        self.last_updated = 0

    def funding_update(self, funding_rate, predicted_funding_rate, next_funding_time, timestamp):
        if funding_rate and next_funding_time:
            self.funding_rate = funding_rate
            self.predicted_funding_rate = predicted_funding_rate
            self.next_update = next_funding_time
            self.last_updated = timestamp



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library

from octobot_trading.util import initializable
from octobot_trading.util.initializable import (
    Initializable,
)

from octobot_trading.util import simulator_updater_utils
from octobot_trading.util import config_util
from octobot_trading.util import databases_util

from octobot_trading.util.simulator_updater_utils import (
    stop_and_pause,
    pause_time_consumer,
    resume_time_consumer,
)
from octobot_trading.util.config_util import (
    is_trader_enabled,
    is_trader_simulator_enabled,
    is_trade_history_loading_enabled,
    is_currency_enabled,
    get_symbols,
    get_all_currencies,
    get_pairs,
    get_market_pair,
    get_reference_market,
    get_traded_pairs_by_currency,
)
from octobot_trading.util.databases_util import (
    get_run_databases_identifier,
)

__all__ = [
    "stop_and_pause",
    "pause_time_consumer",
    "resume_time_consumer",
    "Initializable",
    "is_trader_enabled",
    "is_trader_simulator_enabled",
    "is_trade_history_loading_enabled",
    "is_currency_enabled",
    "get_symbols",
    "get_all_currencies",
    "get_pairs",
    "get_market_pair",
    "get_reference_market",
    "get_traded_pairs_by_currency",
    "get_run_databases_identifier",
]



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.exchanges.abstract_websocket_exchange as abstract_websocket


cdef class WebSocketExchange(abstract_websocket.AbstractWebsocketExchange):
    cdef public str exchange_name

    cdef public list websocket_connectors_tasks
    cdef public list websocket_connectors

    cdef public object websocket_connectors_executors
    cdef public object websocket_connector

    cdef public dict open_sockets_keys
    cdef public dict handled_feeds

    cdef public bint is_websocket_running
    cdef public bint is_websocket_authenticated

    cdef public object restart_task

    # public
    cpdef bint is_handling(self, str feed_name)
    cpdef bint is_feed_available(self, object feed)
    cpdef bint is_feed_requiring_init(self, object feed)
    cpdef void create_feeds(self)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os

import pytest
from math import nan
from octobot_commons.enums import PriceIndexes

from octobot_trading.exchange_data.kline.kline_manager import KlineManager
from tests.test_utils.random_numbers import random_kline, random_price, random_timestamp, random_quantity
from tests import event_loop

pytestmark = pytest.mark.asyncio


@pytest.fixture()
async def kline_manager():
    k_manager = KlineManager()
    await k_manager.initialize()
    return k_manager


async def test_init(kline_manager):
    assert not kline_manager.kline == []


async def test_reset_kline(kline_manager):
    kline_manager.kline = random_kline()
    assert len(kline_manager.kline) == 6
    if not os.getenv('CYTHON_IGNORE'):
        kline_manager._reset_kline()
        assert kline_manager.kline == [nan] * len(PriceIndexes)


async def test_kline_update(kline_manager):
    rc_1 = random_kline()
    kline_manager.kline_update(rc_1)
    assert kline_manager.kline[PriceIndexes.IND_PRICE_CLOSE.value] == rc_1[PriceIndexes.IND_PRICE_CLOSE.value]

    # trigger a new candle
    rc_2 = [0] * len(PriceIndexes)
    rc_2[PriceIndexes.IND_PRICE_CLOSE.value] = 0
    rc_2[PriceIndexes.IND_PRICE_OPEN.value] = 0
    rc_2[PriceIndexes.IND_PRICE_HIGH.value] = 0
    rc_2[PriceIndexes.IND_PRICE_LOW.value] = 0
    rc_2[PriceIndexes.IND_PRICE_VOL.value] = 0
    rc_2[PriceIndexes.IND_PRICE_TIME.value] = random_timestamp()
    kline_manager.kline_update(rc_2)
    assert kline_manager.kline != rc_1

    # don't trigger a new candle
    first_kline = [0] * len(PriceIndexes)
    first_kline[PriceIndexes.IND_PRICE_CLOSE.value] = 0
    first_kline[PriceIndexes.IND_PRICE_OPEN.value] = 0
    first_kline[PriceIndexes.IND_PRICE_HIGH.value] = 0
    first_kline[PriceIndexes.IND_PRICE_LOW.value] = 0
    first_kline[PriceIndexes.IND_PRICE_VOL.value] = 0
    first_kline[PriceIndexes.IND_PRICE_TIME.value] = rc_2[PriceIndexes.IND_PRICE_TIME.value]
    assert kline_manager.kline == first_kline

    # shouldn't use new low
    rc_3 = rc_2
    rc_3[PriceIndexes.IND_PRICE_LOW.value] = random_price(1)
    kline_manager.kline_update(rc_3)
    assert kline_manager.kline == first_kline

    # should use new low
    rc_4 = rc_3
    new_high = random_price(10)
    rc_4[PriceIndexes.IND_PRICE_HIGH.value] = new_high
    kline_manager.kline_update(rc_4)
    second_kline = first_kline
    second_kline[PriceIndexes.IND_PRICE_HIGH.value] = new_high
    assert kline_manager.kline == second_kline

    # shouldn't use new low
    rc_5 = rc_4
    rc_5[PriceIndexes.IND_PRICE_HIGH.value] = new_high - 1
    kline_manager.kline_update(rc_5)
    assert kline_manager.kline == second_kline

    # should use new low
    rc_6 = rc_5
    rc_6[PriceIndexes.IND_PRICE_HIGH.value] = new_high + 1
    third_kline = second_kline
    third_kline[PriceIndexes.IND_PRICE_HIGH.value] = new_high + 1
    kline_manager.kline_update(rc_6)
    assert kline_manager.kline == third_kline

    # should use new vol
    rc_7 = rc_6
    new_vol = random_quantity()
    rc_7[PriceIndexes.IND_PRICE_VOL.value] = new_vol
    kline_4 = third_kline
    kline_4[PriceIndexes.IND_PRICE_VOL.value] = new_vol
    kline_manager.kline_update(rc_7)
    assert kline_manager.kline == kline_4

    # should use new close
    rc_8 = rc_7
    new_price = random_price()
    rc_8[PriceIndexes.IND_PRICE_CLOSE.value] = new_price
    kline_5 = kline_4
    kline_5[PriceIndexes.IND_PRICE_CLOSE.value] = new_price
    kline_manager.kline_update(rc_8)
    assert kline_manager.kline == kline_5

    # shouldn't use new open
    rc_9 = rc_8
    new_open = random_price()
    rc_9[PriceIndexes.IND_PRICE_OPEN.value] = new_open
    kline_manager.kline_update(rc_9)
    assert kline_manager.kline == kline_5



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.constants as constants
import octobot_trading.exchange_data.prices.channel.prices_updater as prices_updater
import octobot_trading.exchange_channel as exchanges_channel


class MarkPriceUpdaterSimulator(prices_updater.MarkPriceUpdater):
    def __init__(self, channel, importer):
        super().__init__(channel)
        self.exchange_data_importer = importer

    async def start(self):
        exchange = self.channel.exchange_manager.exchange
        available_data = exchange.get_real_available_data(exchange.exchange_importers)
        real_data_for_recent_trades = exchange.handles_real_data_for_updater(constants.RECENT_TRADES_CHANNEL,
                                                                             available_data)
        real_data_for_ticker = exchange.handles_real_data_for_updater(constants.TICKER_CHANNEL, available_data)
        # if recent trades and ticker channels are both generated from ohlcv, do not watch them both,
        # prefer ticker
        if real_data_for_ticker or not (real_data_for_recent_trades or real_data_for_ticker):
            await exchanges_channel.get_chan(constants.TICKER_CHANNEL,
                                             self.channel.exchange_manager.id).new_consumer(self.handle_ticker_update)
        if real_data_for_recent_trades:
            await exchanges_channel.get_chan(constants.RECENT_TRADES_CHANNEL, self.channel.exchange_manager.id) \
                .new_consumer(self.handle_recent_trades_update)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import decimal

import octobot_commons.logging as logging
import octobot_commons.constants

import octobot_trading.personal_data.orders.order_factory as order_factory
import octobot_trading.personal_data.orders.order_util as order_util
import octobot_trading.personal_data.orders.decimal_order_adapter as decimal_order_adapter
import octobot_trading.personal_data.trades.trade_factory as trade_factory
import octobot_trading.enums as enums
import octobot_trading.constants
import octobot_trading.errors as errors
import octobot_trading.util as util


class Trader(util.Initializable):
    NO_HISTORY_MESSAGE = "Starting a fresh new trading session using the current portfolio as a profitability " \
                         "reference."

    def __init__(self, config, exchange_manager):
        super().__init__()
        self.exchange_manager = exchange_manager
        self.config = config

        self.risk = octobot_trading.constants.ZERO
        try:
            self.set_risk(decimal.Decimal(str(self.config[octobot_commons.constants.CONFIG_TRADING]
                          [octobot_commons.constants.CONFIG_TRADER_RISK])))
        except KeyError:
            self.set_risk(octobot_trading.constants.ZERO)
        self.allow_artificial_orders = self.config.get(octobot_commons.constants.CONFIG_TRADER_ALLOW_ARTIFICIAL_ORDERS,
                                                       True)

        # logging
        self.trader_type_str = octobot_trading.constants.REAL_TRADER_STR
        self.logger = logging.get_logger(f"{self.__class__.__name__}[{self.exchange_manager.exchange_name}]")

        if not hasattr(self, 'simulate'):
            self.simulate = False
        self.is_enabled = self.__class__.enabled(self.config)

    async def initialize_impl(self):
        self.is_enabled = self.is_enabled and self.exchange_manager.is_trading
        if self.is_enabled:
            await self.exchange_manager.register_trader(self)
        self.logger.debug(f"{'Enabled' if self.is_enabled else 'Disabled'} on {self.exchange_manager.exchange_name}")

    def clear(self):
        self.exchange_manager = None

    @classmethod
    def enabled(cls, config):
        return util.is_trader_enabled(config)

    def set_risk(self, risk):
        min_risk = decimal.Decimal(str(octobot_commons.constants.CONFIG_TRADER_RISK_MIN))
        max_risk = decimal.Decimal(str(octobot_commons.constants.CONFIG_TRADER_RISK_MAX))
        if risk < min_risk:
            self.risk = min_risk
        elif risk > max_risk:
            self.risk = max_risk
        else:
            self.risk = risk
        return self.risk

    """
    Orders
    """

    async def create_order(self, order, loaded: bool = False, params: dict = None, pre_init_callback=None):
        """
        Create a new order from an OrderFactory created order, update portfolio, registers order in order manager and
        notifies order channel.
        :param order: Order to create
        :param loaded: True if this order is fetched from an exchange only and therefore not created by this OctoBot
        :param params: Additional parameters to give to the order upon creation (used in real trading only)
        :param pre_init_callback: A callback function that will be called just before initializing the order
        :return: The crated order instance
        """
        new_order: object = order

        if loaded:
            new_order.is_from_this_octobot = False
            self.logger.debug(f"Order loaded : {new_order.to_string()} ")
        else:
            try:
                params = params or {}
                new_order = await self._create_new_order(new_order, params)
                self.logger.debug(f"Order creation : {new_order.to_string()} ")
            except TypeError as e:
                self.logger.error(f"Fail to create not loaded order : {e}")
                return None

        if pre_init_callback is not None:
            await pre_init_callback(new_order)

        # force initialize to always create open state
        await new_order.initialize()
        return new_order

    async def create_artificial_order(self, order_type, symbol, current_price, quantity, price):
        """
        Creates an OctoBot managed order (managed orders example: stop loss that is not published on the exchange and
        that is maintained internally).
        """
        await self.create_order(order_factory.create_order_instance(trader=self,
                                                                    order_type=order_type,
                                                                    symbol=symbol,
                                                                    current_price=current_price,
                                                                    quantity=quantity,
                                                                    price=price))

    async def edit_order(self, order: object,
                         edited_quantity: decimal.Decimal = None,
                         edited_price: decimal.Decimal = None,
                         edited_stop_price: decimal.Decimal = None,
                         edited_current_price: decimal.Decimal = None,
                         params: dict = None) -> bool:
        """
        Edits an order, might be a simulated or a real order.
        Fields that can be edited are:
            quantity, price, stop_price and current_price
        Portfolio is updated within this call
        :return: True when an order field got updated
        """
        if not order.can_be_edited():
            raise errors.OrderEditError(f"Order can't be edited, order: {order}")
        changed = False
        previous_order_id = order.order_id
        try:
            async with order.lock:
                # now that we got the lock, ensure we can edit the order
                if not self.simulate and not order.is_self_managed() and order.state is not None:
                    # careful here: make sure we are not editing an order on exchange that is being updated
                    # somewhere else
                    async with order.state.refresh_operation():
                        order_params = self.exchange_manager.exchange.get_order_additional_params(order)
                        order_params.update(params or {})
                        # fill in every param as some exchange rely on re-creating the order altogether
                        edited_order = await self.exchange_manager.exchange.edit_order(
                            order.order_id,
                            order.order_type,
                            order.symbol,
                            quantity=order.origin_quantity if edited_quantity is None else edited_quantity,
                            price=order.origin_price if edited_price is None else edited_price,
                            stop_price=edited_stop_price,
                            side=order.side,
                            current_price=edited_current_price,
                            params=order_params
                        )
                        # apply new values from returned order (even order id might have changed)
                        self.logger.debug(f"Successful order edit on {self.exchange_manager.exchange_name}: "
                                          f"{edited_order}")
                        changed = order.update_from_raw(edited_order)
                        # update portfolio from exchange
                        await self.exchange_manager.exchange_personal_data.handle_portfolio_update_from_order(order)
                else:
                    # consider order as cancelled to release portfolio amounts
                    self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.update_portfolio_available(
                        order, is_new_order=False
                    )
                    changed = order.update(
                        order.symbol,
                        quantity=edited_quantity,
                        price=edited_stop_price if edited_price is None else edited_price,
                        stop_price=edited_stop_price,
                        current_price=edited_current_price,
                    )
                    # consider order as new order to lock portfolio amounts
                    self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.update_portfolio_available(
                        order, is_new_order=True
                    )
                self.logger.info(f"Edited order: {order}")
            return changed
        finally:
            if previous_order_id != order.order_id:
                # order id changed: update orders_manager to keep consistency
                self.exchange_manager.exchange_personal_data.orders_manager.replace_order(previous_order_id, order)

    async def _create_new_order(self, new_order: object, params: dict) -> object:
        """
        Creates an exchange managed order, it might be a simulated or a real order.
        Portfolio will be updated by the create order state after order will be initialized
        """
        updated_order = new_order
        if not self.simulate and not new_order.is_self_managed():
            order_params = self.exchange_manager.exchange.get_order_additional_params(new_order)
            order_params.update(params)
            created_order = await self.exchange_manager.exchange.create_order(new_order.order_type,
                                                                              new_order.symbol,
                                                                              new_order.origin_quantity,
                                                                              new_order.origin_price,
                                                                              new_order.origin_stop_price,
                                                                              new_order.side,
                                                                              new_order.created_last_price,
                                                                              params=order_params)

            self.logger.info(f"Created order on {self.exchange_manager.exchange_name}: {created_order}")

            # get real order from exchange
            updated_order = order_factory.create_order_instance_from_raw(self, created_order, force_open=True)

            # rebind local elements to new order instance
            if new_order.order_group:
                updated_order.add_to_order_group(new_order.order_group)
            updated_order.tag = new_order.tag
            updated_order.chained_orders = new_order.chained_orders
            for chained_order in new_order.chained_orders:
                chained_order.triggered_by = updated_order
            updated_order.triggered_by = new_order.triggered_by
            updated_order.has_been_bundled = new_order.has_been_bundled
            updated_order.exchange_creation_params = new_order.exchange_creation_params
            updated_order.is_waiting_for_chained_trigger = new_order.is_waiting_for_chained_trigger
            updated_order.set_shared_signal_order_id(new_order.shared_signal_order_id)
        return updated_order

    async def bundle_chained_order_with_uncreated_order(self, order, chained_order, **kwargs):
        """
        Creates and bundles an order as a chained order to the given order.
        When supported and in real trading, return the stop loss parameters to be given when
        pushing the initial order on exchange
        :param order: the order to create a chained order from after fill
        :param chained_order: the chained order to create when the 1st order is filled
        :return: parameters with chained order details if supported
        """
        params = {}
        is_bundled = self.exchange_manager.exchange.supports_bundled_order_on_order_creation(
            order, chained_order.order_type
        )
        if is_bundled:
            if chained_order.order_type is enums.TraderOrderType.STOP_LOSS:
                params.update(self.exchange_manager.exchange.get_bundled_order_parameters(
                    stop_loss_price=chained_order.origin_price
                ))
            if chained_order.order_type in (enums.TraderOrderType.BUY_MARKET, enums.TraderOrderType.SELL_MARKET,
                                            enums.TraderOrderType.BUY_LIMIT, enums.TraderOrderType.SELL_LIMIT):
                params.update(self.exchange_manager.exchange.get_bundled_order_parameters(
                    take_profit_price=chained_order.origin_price
                ))
        await chained_order.set_as_chained_order(order, is_bundled, {}, **kwargs)
        order.add_chained_order(chained_order)
        return params

    async def cancel_order(self, order: object, ignored_order: object = None) -> bool:
        """
        Cancels the given order and updates the portfolio, publish in order channel
        if order is from a real exchange.
        :param order: Order to cancel
        :param ignored_order: Order not to cancel if found in groupped orders recursive cancels (ex: avoid cancelling
        a filled order)
        :return: None
        """
        if order and order.is_open():
            # always cancel this order first to avoid infinite loop followed by deadlock
            return await self._handle_order_cancellation(order, ignored_order)
        return False

    async def _handle_order_cancellation(self, order: object, ignored_order: object) -> bool:
        success = True
        async with order.lock:
            if order.is_waiting_for_chained_trigger:
                # order will just never get created
                order.is_waiting_for_chained_trigger = False
                return success
            # if real order: cancel on exchange
            if not self.simulate and not order.is_self_managed():
                success = await self.exchange_manager.exchange.cancel_order(order.order_id, order.symbol)
                if not success:
                    # retry to cancel order
                    success = await self.exchange_manager.exchange.cancel_order(order.order_id, order.symbol)
                if not success:
                    self.logger.warning(f"Failed to cancel order {order}")
                    return False
                else:
                    order.status = octobot_trading.enums.OrderStatus.CLOSED
                    self.logger.debug(f"Successfully cancelled order {order}")
            else:
                order.status = octobot_trading.enums.OrderStatus.CANCELED

        # call CancelState termination
        await order.on_cancel(force_cancel=success,
                              is_from_exchange_data=False,
                              ignored_order=ignored_order)
        return True

    async def cancel_order_with_id(self, order_id):
        """
        Gets order matching order_id from the OrderManager and calls self.cancel_order() on it
        :param order_id: Id of the order to cancel
        :return: True if cancel is successful, False if order is not found or cancellation failed
        """
        try:
            return await self.cancel_order(
                self.exchange_manager.exchange_personal_data.orders_manager.get_order(order_id)
            )
        except KeyError:
            return False

    async def cancel_open_orders(self, symbol, cancel_loaded_orders=True, side=None) -> (bool, list):
        """
        Should be called only if the goal is to cancel all open orders for a given symbol
        :param symbol: The symbol to cancel all orders on
        :param cancel_loaded_orders: When True, also cancels loaded orders (order that are not from this bot instance)
        :param side: When set, only cancels orders from this side
        :return: (True, orders): True if all orders got cancelled, False if an error occurred and the list of
        cancelled orders
        """
        all_cancelled = True
        cancelled_orders = []
        for order in self.exchange_manager.exchange_personal_data.orders_manager.get_open_orders():
            if order.symbol == symbol and \
                    (side is None or order.side is side) and \
                    not order.is_cancelled() and \
                    (cancel_loaded_orders or order.is_from_this_octobot):
                cancelled = await self.cancel_order(order)
                if cancelled:
                    cancelled_orders.append(order)
                all_cancelled = cancelled and all_cancelled
        return all_cancelled, cancelled_orders

    async def cancel_all_open_orders_with_currency(self, currency) -> bool:
        """
        Should be called only if the goal is to cancel all open orders for each traded symbol containing the
        given currency.
        :param currency: Currency to find trading pairs to cancel orders on.
        :return: True if all orders got cancelled, False if an error occurred
        """
        all_cancelled = True
        symbols = util.get_pairs(self.config, currency, enabled_only=True)
        if symbols:
            for symbol in symbols:
                all_cancelled = (await self.cancel_open_orders(symbol))[0] and all_cancelled
        return all_cancelled

    async def cancel_all_open_orders(self) -> bool:
        """
        Cancel all open orders registered on this bot.
        :return: True if all orders got cancelled, False if an error occurred
        """
        all_cancelled = True
        for order in self.exchange_manager.exchange_personal_data.orders_manager.get_open_orders():
            if not order.is_cancelled():
                all_cancelled = await self.cancel_order(order) and all_cancelled
        return all_cancelled

    async def _sell_everything(self, symbol, inverted, timeout=None):
        created_orders = []
        order_type = octobot_trading.enums.TraderOrderType.BUY_MARKET \
            if inverted else octobot_trading.enums.TraderOrderType.SELL_MARKET
        async with self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.lock:
            current_symbol_holding, current_market_quantity, _, price, symbol_market = \
                await order_util.get_pre_order_data(self.exchange_manager, symbol, timeout=timeout)
            if inverted:
                if price > 0:
                    quantity = current_market_quantity / price
                else:
                    quantity = 0
            else:
                quantity = current_symbol_holding
            for order_quantity, order_price in decimal_order_adapter.decimal_check_and_adapt_order_details_if_necessary(quantity, price,
                                                                                                        symbol_market):
                current_order = order_factory.create_order_instance(trader=self,
                                                                    order_type=order_type,
                                                                    symbol=symbol,
                                                                    current_price=order_price,
                                                                    quantity=order_quantity,
                                                                    price=order_price)
                created_orders.append(
                    await self.create_order(current_order))
        return created_orders

    async def sell_all(self, currencies_to_sell=None, timeout=None):
        """
        Sell every currency in portfolio for reference market using market orders.
        :param currencies_to_sell: List of currencies to sell, default values consider every currency in portfolio
        :param timeout: Timeout to get market price
        :return: The created orders
        """
        orders = []
        currency_list = self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.portfolio

        if not currencies_to_sell:
            currencies = currency_list
        else:
            currencies = [currency
                          for currency in currencies_to_sell
                          if currency in currency_list]

        for currency in currencies:
            symbol, inverted = util.get_market_pair(self.config, currency, enabled_only=True)
            if symbol:
                orders += await self._sell_everything(symbol, inverted, timeout=timeout)
        return orders

    def parse_order_id(self, order_id):
        return order_id

    def convert_order_to_trade(self, order):
        """
        Convert an order instance to Trade
        :return: the new Trade instance from order
        """
        return trade_factory.create_trade_from_order(order)

    """
    Positions
    """

    async def close_position(self, position, limit_price=None, timeout=1):
        """
        Creates a close position order
        :param position: the position to close
        :param limit_price: the close order limit price if None uses a market order
        :param timeout: the mark price timeout
        :return: the list of created orders
        """
        created_orders = []
        async with self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.lock:
            _, _, _, price, symbol_market = await order_util.get_pre_order_data(self.exchange_manager,
                                                                                position.symbol,
                                                                                timeout=timeout)
            for order_quantity, order_price in decimal_order_adapter.decimal_check_and_adapt_order_details_if_necessary(
                    position.get_quantity_to_close().copy_abs(), price, symbol_market):

                if limit_price is not None:
                    order_type = enums.TraderOrderType.SELL_LIMIT \
                        if position.is_long() else enums.TraderOrderType.BUY_LIMIT
                else:
                    order_type = enums.TraderOrderType.SELL_MARKET \
                        if position.is_long() else enums.TraderOrderType.BUY_MARKET

                # TODO add reduce_only or close_position attribute
                current_order = order_factory.create_order_instance(trader=self,
                                                                    order_type=order_type,
                                                                    symbol=position.symbol,
                                                                    current_price=order_price,
                                                                    quantity=order_quantity,
                                                                    price=limit_price
                                                                    if limit_price is not None else order_price)
                created_orders.append(
                    await self.create_order(current_order))
        return created_orders

    async def withdraw(self, amount, currency):
        """
        Removes the given amount from the portfolio. Only works in simulated portfolios
        :param amount: the amount to withdraw
        :param currency: the currency to withdraw
        """
        async with self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.lock:
            await self.exchange_manager.exchange_personal_data.handle_portfolio_update_from_withdrawal(amount, currency)

    async def set_leverage(self, symbol, side, leverage):
        """
        Updates the symbol contract leverage
        Can raise InvalidLeverageValue if leverage value is not matching requirements
        :param symbol: the symbol to update
        :param side: the side to update (TODO)
        :param leverage: the new leverage value
        """
        contract = self.exchange_manager.exchange.get_pair_future_contract(symbol)
        if not contract.check_leverage_update(leverage):
            raise errors.InvalidLeverageValue(f"Trying to update leverage with {leverage} "
                                              f"but maximal value is {contract.maximum_leverage}")
        if contract.current_leverage != leverage:
            if not self.simulate:
                await self.exchange_manager.exchange.set_symbol_leverage(
                    symbol=symbol,
                    leverage=leverage
                )
            contract.set_current_leverage(leverage)

    async def set_symbol_take_profit_stop_loss_mode(self, symbol, new_mode: enums.TakeProfitStopLossMode):
        """
        Updates the take profit and stop loss mode for the given symbol
        Raises NotImplementedError if the endpoint is not implemented on exchange
        :param symbol: the symbol to update
        :param new_mode: the take_profit_stop_loss_mode value
        """
        contract = self.exchange_manager.exchange.get_pair_future_contract(symbol)
        if contract.take_profit_stop_loss_mode != new_mode:
            if not self.simulate:
                await self.exchange_manager.exchange.set_symbol_partial_take_profit_stop_loss(
                    symbol, contract.is_inverse_contract(), new_mode)
            contract.set_take_profit_stop_loss_mode(new_mode)

    async def set_margin_type(self, symbol, side, margin_type):
        """
        Updates the symbol contract margin type
        TODO: recreate position instances if any
        :param symbol: the symbol to update
        :param side: the side to update (TODO)
        :param margin_type: the new margin type (enums.MarginType)
        """
        contract = self.exchange_manager.exchange.get_pair_future_contract(symbol)
        if not self.simulate:
            await self.exchange_manager.exchange.set_symbol_margin_type(
                symbol=symbol,
                isolated=margin_type is enums.MarginType.ISOLATED
            )
        contract.set_margin_type(
            is_isolated=margin_type is enums.MarginType.ISOLATED,
            is_cross=margin_type is enums.MarginType.CROSS
        )

    async def set_position_mode(self, symbol, position_mode):
        """
        Updates the symbol contract position mode
        :param symbol: the symbol to update
        :param position_mode: the new position mode (enums.PositionMode)
        """
        if self._has_open_position(symbol):
            raise errors.TooManyOpenPositionError("Can't update position mode when having open position.")
        contract = self.exchange_manager.exchange.get_pair_future_contract(symbol)
        if not self.simulate:
            await self.exchange_manager.exchange.set_symbol_position_mode(
                symbol=symbol,
                one_way=position_mode is enums.PositionMode.ONE_WAY
            )
        contract.set_position_mode(
            is_one_way=position_mode is enums.PositionMode.ONE_WAY,
            is_hedge=position_mode is enums.PositionMode.HEDGE
        )

    def _has_open_position(self, symbol):
        """
        Checks if open position exists for :symbol:
        :param symbol: the position symbol
        :return: True if open position for :symbol: exists
        """
        return len(self.exchange_manager.exchange_personal_data.positions_manager.get_symbol_positions(
            symbol=symbol)) != 0



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal

import octobot_commons.logging as logging

import octobot_trading.util as util
import octobot_trading.constants as constants

LOGGER = logging.get_logger(constants.API_LOGGER_TAG)


def get_trader(exchange_manager):
    return exchange_manager.trader


def has_trader(exchange_manager):
    return exchange_manager.trader is not None


def is_trader_enabled_in_config_from_exchange_manager(exchange_manager) -> bool:
    return exchange_manager.trader.enabled(exchange_manager.config)


def is_trader_existing_and_enabled(exchange_manager) -> bool:
    return False if exchange_manager.trader is None else exchange_manager.trader.is_enabled


def is_trader_enabled(exchange_manager) -> bool:
    return exchange_manager.trader.is_enabled


def is_trader_enabled_in_config(config) -> bool:
    return util.is_trader_enabled(config)


def is_trader_simulator_enabled_in_config(config) -> bool:
    return util.is_trader_simulator_enabled(config)


def set_trading_enabled(exchange_manager, enabled) -> None:
    exchange_manager.trader.is_enabled = enabled


def is_trader_simulated(exchange_manager) -> bool:
    return exchange_manager.is_trader_simulated


def get_trader_risk(exchange_manager) -> float:
    return exchange_manager.trader.risk


def set_trader_risk(exchange_manager, risk: decimal.Decimal) -> float:
    return exchange_manager.trader.set_risk(decimal.Decimal(risk))


async def sell_all_everything_for_reference_market(exchange_manager) -> list:
    return await exchange_manager.trader.sell_all()


async def sell_currency_for_reference_market(exchange_manager, currency) -> list:
    return await exchange_manager.trader.sell_all([currency])



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.modes.script_keywords.basic_keywords.user_inputs import (
    user_input,
    save_user_input,
    get_user_inputs,
    clear_user_inputs,
    get_activation_topics,
)

from octobot_trading.modes.script_keywords.basic_keywords.configuration import (
    user_select_leverage,
    user_select_emit_trading_signals,
    set_leverage,
    set_partial_take_profit_stop_loss,
)

from octobot_trading.modes.script_keywords.basic_keywords.trading_signals import (
    is_emitting_trading_signals,
    emit_trading_signals,
)


from octobot_trading.modes.script_keywords.basic_keywords.run_persistence import (
    set_plot_orders,
    store_orders,
    store_trade,
    store_transactions,
    save_metadata,
    save_portfolio,
    clear_run_data,
    clear_orders_cache,
    clear_trades_cache,
    clear_transactions_cache,
    clear_all_tables,
)


__all__ = [
    "user_input",
    "save_user_input",
    "get_user_inputs",
    "clear_user_inputs",
    "get_activation_topics",
    "user_select_leverage",
    "user_select_emit_trading_signals",
    "is_emitting_trading_signals",
    "emit_trading_signals",
    "set_leverage",
    "set_partial_take_profit_stop_loss",
    "set_plot_orders",
    "store_orders",
    "store_trade",
    "store_transactions",
    "save_metadata",
    "save_portfolio",
    "clear_run_data",
    "clear_orders_cache",
    "clear_trades_cache",
    "clear_transactions_cache",
    "clear_all_tables",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import os
import pytest

import octobot_commons.constants as commons_constants
from octobot_backtesting.backtesting import Backtesting
from octobot_backtesting.constants import CONFIG_BACKTESTING
import octobot_backtesting.time as backtesting_time
from octobot_commons.asyncio_tools import wait_asyncio_next_cycle
from octobot_commons.enums import TimeFrames

from octobot_commons.tests.test_config import load_test_config
from octobot_trading.api.exchange import create_exchange_builder, cancel_ccxt_throttle_task
from octobot_trading.exchanges.exchange_manager import ExchangeManager
from octobot_trading.exchanges.traders.trader_simulator import TraderSimulator

pytestmark = pytest.mark.asyncio

TESTS_FOLDER = "tests"
TESTS_STATIC_FOLDER = os.path.join(TESTS_FOLDER, "static")
DEFAULT_EXCHANGE_NAME = "binance"


@pytest.fixture
async def exchange_manager():
    exchange_manager_instance = ExchangeManager(load_test_config(), DEFAULT_EXCHANGE_NAME)
    exchange_manager_instance.is_spot_only = True
    exchange_manager_instance.is_simulated = False
    await exchange_manager_instance.initialize()
    yield exchange_manager_instance
    cancel_ccxt_throttle_task()
    await exchange_manager_instance.stop()
    # let updaters gracefully shutdown
    await wait_asyncio_next_cycle()


@pytest.fixture
async def simulated_exchange_manager():
    exchange_manager_instance = ExchangeManager(load_test_config(), DEFAULT_EXCHANGE_NAME)
    exchange_manager_instance.is_spot_only = True
    exchange_manager_instance.is_simulated = True
    await exchange_manager_instance.initialize()
    yield exchange_manager_instance
    cancel_ccxt_throttle_task()
    await exchange_manager_instance.stop()
    # let updaters gracefully shutdown
    await wait_asyncio_next_cycle()


@pytest.fixture
async def margin_exchange_manager():
    exchange_manager_instance = ExchangeManager(load_test_config(), DEFAULT_EXCHANGE_NAME)
    exchange_manager_instance.is_spot_only = False
    exchange_manager_instance.is_margin = True
    await exchange_manager_instance.initialize()
    yield exchange_manager_instance
    cancel_ccxt_throttle_task()
    await exchange_manager_instance.stop()
    # let updaters gracefully shutdown
    await wait_asyncio_next_cycle()


@pytest.fixture
async def margin_simulated_exchange_manager():
    exchange_manager_instance = ExchangeManager(load_test_config(), DEFAULT_EXCHANGE_NAME)
    exchange_manager_instance.is_spot_only = False
    exchange_manager_instance.is_simulated = True
    exchange_manager_instance.is_margin = True
    await exchange_manager_instance.initialize()
    yield exchange_manager_instance
    cancel_ccxt_throttle_task()
    await exchange_manager_instance.stop()
    # let updaters gracefully shutdown
    await wait_asyncio_next_cycle()


@pytest.fixture
async def future_exchange_manager():
    exchange_manager_instance = ExchangeManager(load_test_config(), DEFAULT_EXCHANGE_NAME)
    exchange_manager_instance.is_spot_only = False
    exchange_manager_instance.is_future = True
    await exchange_manager_instance.initialize()
    yield exchange_manager_instance
    cancel_ccxt_throttle_task()
    await exchange_manager_instance.stop()
    # let updaters gracefully shutdown
    await wait_asyncio_next_cycle()


@pytest.fixture
async def future_simulated_exchange_manager():
    exchange_manager_instance = ExchangeManager(load_test_config(), DEFAULT_EXCHANGE_NAME)
    exchange_manager_instance.is_spot_only = False
    exchange_manager_instance.is_simulated = True
    exchange_manager_instance.is_future = True
    await exchange_manager_instance.initialize()
    yield exchange_manager_instance
    cancel_ccxt_throttle_task()
    await exchange_manager_instance.stop()
    # let updaters gracefully shutdown
    await wait_asyncio_next_cycle()


@pytest.fixture
async def exchange_builder(request):
    config = None
    exchange_name = DEFAULT_EXCHANGE_NAME
    if hasattr(request, "param"):
        config, exchange_name = request.param

    exchange_builder_instance = create_exchange_builder(config if config is not None else load_test_config(),
                                                        exchange_name).is_simulated().is_rest_only()
    yield exchange_builder_instance
    await exchange_builder_instance.exchange_manager.stop()


# SIMULATED / BACKTESTING
DEFAULT_BACKTESTING_SYMBOL = "BTC/USDT"
DEFAULT_BACKTESTING_SPLIT_SYMBOL = "BTC", "USDT"
DEFAULT_BACKTESTING_CURRENCY = "BTC"
DEFAULT_BACKTESTING_MARKET = "USDT"
DEFAULT_BACKTESTING_TF = TimeFrames.ONE_HOUR


@pytest.fixture
async def backtesting_config(request):
    config = load_test_config()
    config[CONFIG_BACKTESTING] = {}
    config[CONFIG_BACKTESTING][commons_constants.CONFIG_ENABLED_OPTION] = True
    if hasattr(request, "param"):
        ref_market = request.param
        config[commons_constants.CONFIG_TRADING][commons_constants.CONFIG_TRADER_REFERENCE_MARKET] = ref_market
    return config


@pytest.fixture
async def fake_backtesting(backtesting_config):
    return Backtesting(config=backtesting_config,
                       exchange_ids=[],
                       matrix_id="",
                       backtesting_files=[])


@pytest.fixture
async def backtesting_exchange_manager(request, backtesting_config, fake_backtesting):
    config = None
    exchange_name = DEFAULT_EXCHANGE_NAME
    is_spot = True
    is_margin = False
    is_future = False
    if hasattr(request, "param"):
        config, exchange_name, is_spot, is_margin, is_future = request.param

    if config is None:
        config = backtesting_config
    exchange_manager_instance = ExchangeManager(config, exchange_name)
    exchange_manager_instance.is_backtesting = True
    exchange_manager_instance.is_spot_only = is_spot
    exchange_manager_instance.is_margin = is_margin
    exchange_manager_instance.is_future = is_future
    exchange_manager_instance.backtesting = fake_backtesting
    exchange_manager_instance.backtesting.time_manager = backtesting_time.TimeManager(config)
    await exchange_manager_instance.initialize()
    yield exchange_manager_instance
    await exchange_manager_instance.stop()


@pytest.fixture
async def simulated_trader(simulated_exchange_manager):
    config = load_test_config()
    trader_instance = TraderSimulator(config, simulated_exchange_manager)
    await trader_instance.initialize()
    return config, simulated_exchange_manager, trader_instance


@pytest.fixture
async def backtesting_trader(backtesting_config, backtesting_exchange_manager):
    trader_instance = TraderSimulator(backtesting_config, backtesting_exchange_manager)
    await trader_instance.initialize()
    return backtesting_config, backtesting_exchange_manager, trader_instance


@pytest.fixture
async def margin_backtesting_trader(backtesting_config, backtesting_exchange_manager):
    trader_instance = TraderSimulator(backtesting_config, backtesting_exchange_manager)
    await trader_instance.initialize()
    return backtesting_config, backtesting_exchange_manager, trader_instance


@pytest.fixture
async def future_backtesting_trader(backtesting_config, backtesting_exchange_manager):
    trader_instance = TraderSimulator(backtesting_config, backtesting_exchange_manager)
    await trader_instance.initialize()
    return backtesting_config, backtesting_exchange_manager, trader_instance



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.symbol_util as symbol_util

import octobot_trading.exchange_channel as exchange_channel
import octobot_trading.modes.channel as modes_channel
import octobot_trading.enums as enums
import octobot_trading.errors as errors
import octobot_trading.constants as constants
from octobot_trading.enums import ExchangeConstantsMarketStatusColumns as Ecmsc


class AbstractTradingModeConsumer(modes_channel.ModeChannelConsumer):
    def __init__(self, trading_mode):
        super().__init__()
        self.trading_mode = trading_mode
        self.exchange_manager = trading_mode.exchange_manager

    def flush(self):
        self.trading_mode = None
        self.exchange_manager = None

    async def internal_callback(self, trading_mode_name, cryptocurrency, symbol, time_frame, final_note, state, data):
        # creates a new order (or multiple split orders), always check self.can_create_order() first.
        try:
            await self.create_order_if_possible(symbol, final_note, state, data=data)
        except errors.MissingMinimalExchangeTradeVolume:
            self.logger.info(f"Not enough funds to create a new order: {self.exchange_manager.exchange_name} "
                             f"exchange minimal order volume has not been reached.")

    async def create_new_orders(self, symbol, final_note, state, **kwargs):
        raise NotImplementedError("create_new_orders is not implemented")

    async def create_order_if_possible(self, symbol, final_note, state, **kwargs) -> list:
        """
        For each trader call the creator to check if order creation is possible and create it.
        Will retry once on failure
        :return: None
        """
        self.logger.debug(f"Entering create_order_if_possible for {symbol}")
        try:
            async with self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio.lock:
                if await self.can_create_order(symbol, state):
                    try:
                        return await self.create_new_orders(symbol, final_note, state, **kwargs)
                    except errors.MissingMinimalExchangeTradeVolume:
                        raise
                    except errors.MissingFunds:
                        try:
                            # second chance: force portfolio update and retry
                            await exchange_channel.get_chan(constants.BALANCE_CHANNEL,
                                                     self.exchange_manager.id).get_internal_producer(). \
                                refresh_real_trader_portfolio(True)

                            return await self.create_new_orders(symbol, final_note, state, **kwargs)
                        except errors.MissingFunds as e:
                            self.logger.error(f"Failed to create order on second attempt : {e})")
                    except Exception as e:
                        self.logger.exception(e, True, f"Error when creating order: {e}")
            self.logger.debug(f"Skipping order creation for {symbol}: not enough available funds")
            return []
        finally:
            self.logger.debug(f"Exiting create_order_if_possible for {symbol}")

    # Can be overwritten
    async def can_create_order(self, symbol, state):
        currency, market = symbol_util.split_symbol(symbol)
        portfolio = self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio

        # get symbol min amount when creating order
        symbol_limit = self.exchange_manager.exchange.get_market_status(symbol)[Ecmsc.LIMITS.value]
        symbol_min_amount = symbol_limit[Ecmsc.LIMITS_AMOUNT.value][Ecmsc.LIMITS_AMOUNT_MIN.value]
        order_min_amount = symbol_limit[Ecmsc.LIMITS_COST.value][Ecmsc.LIMITS_COST_MIN.value]

        if symbol_min_amount is None:
            symbol_min_amount = 0

        # short cases => sell => need this currency
        if state == enums.EvaluatorStates.VERY_SHORT.value or state == enums.EvaluatorStates.SHORT.value:
            return portfolio.get_currency_portfolio(currency).available > symbol_min_amount

        # long cases => buy => need money(aka other currency in the pair) to buy this currency
        elif state == enums.EvaluatorStates.LONG.value or state == enums.EvaluatorStates.VERY_LONG.value:
            return portfolio.get_currency_portfolio(market).available > order_min_amount

        # other cases like neutral state or unfulfilled previous conditions
        return False

    def get_holdings_ratio(self, currency):
        return self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder \
            .get_currency_holding_ratio(currency)

    def get_number_of_traded_assets(self):
        return len(self.exchange_manager.exchange_personal_data.portfolio_manager.portfolio_value_holder
                   .origin_crypto_currencies_values)


def check_factor(min_val, max_val, factor):
    """
    Checks if factor is min_val < factor < max_val
    :param min_val:
    :param max_val:
    :param factor:
    :return:
    """
    if factor > max_val:
        return max_val
    if factor < min_val:
        return min_val
    return factor



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.storage as storage


def init_bot_storage(bot_id, config, tentacles_setup_config):
    if not storage.RunDatabasesProvider.instance().has_bot_id(bot_id):
        # only one run database per bot id
        storage.RunDatabasesProvider.instance().add_bot_id(bot_id, config, tentacles_setup_config)


def get_run_db(bot_id):
    return storage.RunDatabasesProvider.instance().get_run_db(bot_id)


def get_symbol_db(bot_id, exchange, symbol):
    return storage.RunDatabasesProvider.instance().get_symbol_db(bot_id, exchange, symbol)


async def close_bot_storage(bot_id):
    if storage.RunDatabasesProvider.instance().has_bot_id(bot_id):
        await storage.RunDatabasesProvider.instance().close(bot_id)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.exchange_data as exchange_data

cdef class ExchangeSymbolsData:
    cdef public object logger

    cdef public dict exchange_symbol_data
    cdef public dict config

    cdef public object exchange
    cdef public object exchange_manager

    cpdef public exchange_data.ExchangeSymbolData get_exchange_symbol_data(self, str symbol, bint allow_creation=*)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.personal_data.positions as positions_personal_data
cimport octobot_trading.util as util


cdef class PositionsManager(util.Initializable):
    cdef object logger

    cdef public bint positions_initialized

    cdef object trader

    cdef public object positions

    cdef void _reset_positions(self)
    cdef positions_personal_data.Position _get_or_create_position(self, str symbol, object side)
    cdef object _create_symbol_position(self, str symbol, str position_id)
    cdef str _generate_position_id(self, str symbol, object side, object expiration_time=*)
    cdef list _get_symbol_positions(self, str symbol)

    cpdef positions_personal_data.Position get_symbol_position(self, str symbol, object side)
    cpdef positions_personal_data.Position get_order_position(self, object order, object contract=*)
    cpdef list get_symbol_positions(self, str symbol=*)
    cpdef bint upsert_position_instance(self, positions_personal_data.Position position)
    cpdef void clear(self)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.singleton as singleton
import octobot_commons.databases as databases
import octobot_commons.errors as errors
import octobot_trading.util as util


class RunDatabasesProvider(singleton.Singleton):
    def __init__(self):
        self.run_databases = {}

    def add_bot_id(self, bot_id, config, tentacles_setup_config, with_lock=False, cache_size=None):
        run_dbs_identifier = util.get_run_databases_identifier(config, tentacles_setup_config)
        self.run_databases[bot_id] = databases.MetaDatabase(run_dbs_identifier, with_lock=with_lock,
                                                            cache_size=cache_size)

    def has_bot_id(self, bot_id):
        return bot_id in self.run_databases

    def get_run_databases_identifier(self, bot_id):
        return self.run_databases[bot_id].run_dbs_identifier

    def get_run_db(self, bot_id):
        return self.run_databases[bot_id].get_run_db()

    def get_orders_db(self, bot_id, exchange=None):
        return self.run_databases[bot_id].get_orders_db(exchange=exchange)

    def get_trades_db(self, bot_id, exchange=None):
        return self.run_databases[bot_id].get_trades_db(exchange=exchange)

    def get_transactions_db(self, bot_id, exchange=None):
        return self.run_databases[bot_id].get_transactions_db(exchange=exchange)

    def get_backtesting_metadata_db(self, bot_id):
        return self.run_databases[bot_id].get_backtesting_metadata_db()

    def get_symbol_db(self, bot_id, exchange, symbol):
        if not symbol:
            raise errors.DatabaseNotFoundError("symbol parameter has to be provided")
        return self.run_databases[bot_id].get_symbol_db(exchange, symbol)

    async def close(self, bot_id):
        await self.run_databases[bot_id].close()
        self.run_databases.pop(bot_id)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_commons.singleton as singleton

cdef class RunDatabasesProvider(singleton.Singleton):
    cdef public dict run_databases

    cpdef void add_bot_id(self, str bot_id, dict config, object tentacles_setup_config, bint with_lock=*, object cache_size=*)
    cpdef bint has_bot_id(self, str bot_id)
    cpdef object get_run_db(self, str bot_id)
    cpdef object get_orders_db(self, str bot_id, str exchange=*)
    cpdef object get_trades_db(self, str bot_id, str exchange=*)
    cpdef object get_transactions_db(self, str bot_id, str exchange=*)
    cpdef object get_backtesting_metadata_db(self, str bot_id)
    cpdef object get_symbol_db(self, str bot_id, str exchange, str symbol)



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.


cpdef bint is_trader_enabled(dict config)
cpdef bint is_trader_simulator_enabled(dict config)
cpdef bint is_currency_enabled(dict config, str currency, bint default_value)
cpdef bint is_trade_history_loading_enabled(dict config, bint default=*)
cpdef list get_symbols(dict config, bint enabled_only)
cpdef set get_all_currencies(dict config, bint enabled_only=*)
cpdef list get_pairs(dict config, str currency, bint enabled_only=*)
cpdef tuple get_market_pair(dict config, str currency, bint enabled_only=*)
cpdef str get_reference_market(dict config)
cpdef dict get_traded_pairs_by_currency(dict config)

cdef bint _is_trader_enabled(dict config, str trader_key)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

import octobot_trading.constants as constants
from octobot_trading.exchange_data.funding.funding_manager import FundingManager
from tests.test_utils.random_numbers import random_timestamp, random_funding_rate
from tests import event_loop

pytestmark = pytest.mark.asyncio


@pytest.fixture()
async def funding_manager():
    fund_manager = FundingManager()
    await fund_manager.initialize()
    return fund_manager


async def test_init(funding_manager):
    assert funding_manager.next_update == 0
    assert funding_manager.last_updated == 0
    assert funding_manager.funding_rate == constants.ZERO
    assert funding_manager.predicted_funding_rate == constants.ZERO


async def test_reset_funding(funding_manager):
    funding_manager.next_update = random_timestamp()
    funding_manager.last_updated = random_timestamp()
    funding_manager.funding_rate = random_funding_rate()
    funding_manager.predicted_funding_rate = random_funding_rate()
    funding_manager.reset_funding()
    assert funding_manager.next_update == 0
    assert funding_manager.last_updated == 0
    assert funding_manager.funding_rate == constants.ZERO
    assert funding_manager.predicted_funding_rate == constants.ZERO


async def test_funding_update(funding_manager):
    last_updated = random_timestamp()
    next_updated = random_timestamp(last_updated)
    funding_rate = random_funding_rate()
    predicted_funding_rate = random_funding_rate()
    funding_manager.funding_update(funding_rate, predicted_funding_rate, next_updated, last_updated)
    assert funding_manager.next_update == next_updated
    assert funding_manager.last_updated == last_updated
    assert funding_manager.funding_rate == funding_rate
    assert funding_manager.predicted_funding_rate == predicted_funding_rate



# cython: language_level=3
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
cimport octobot_trading.exchange_data.ohlcv.channel.ohlcv as ohlcv_channel


cdef class OHLCVUpdater(ohlcv_channel.OHLCVProducer):
    cdef list tasks

    cdef bint is_initialized
    cdef dict initialized_candles_by_tf_by_symbol

    cdef list _get_traded_pairs(self)
    cdef list _get_time_frames(self)
    cdef int _get_historical_candles_count(self)
    cdef double _ensure_correct_sleep_time(self, double sleep_time_candidate, double time_frame_sleep)
    cdef void _set_initialized(self, str pair, object time_frame, bint initialized)



# pylint: disable=protected-access
#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import asyncio
import trading_backend

import octobot_trading.constants
import octobot_trading.enums
import octobot_trading.exchanges as exchanges
import octobot_trading.exchange_data as exchange_data

import octobot_backtesting.api as backtesting_api


def create_exchange_builder(config, exchange_name: str) -> exchanges.ExchangeBuilder:
    return exchanges.create_exchange_builder_instance(config, exchange_name)


async def stop_exchange(exchange_manager) -> None:
    await exchange_manager.stop()


def get_exchange_configurations_from_exchange_name(exchange_name: str) -> dict:
    return exchanges.Exchanges.instance().get_exchanges(exchange_name)


def get_exchange_manager_from_exchange_name_and_id(exchange_name: str, exchange_id: str) -> object:
    return exchanges.Exchanges.instance().get_exchange(exchange_name, exchange_id).exchange_manager


async def get_exchange_available_time_frames(
        exchange_name: str,
        exchange_lib: octobot_trading.enums.ExchangeWrapperLibs = octobot_trading.enums.ExchangeWrapperLibs.CCXT
) -> list:
    """
    When using CCXT, prefer using the sync lib since no request is required to get time frames
    :param exchange_name: name of the exchange
    :param exchange_lib: octobot_trading.enums.ExchangeWrapperLibs to use
    :return: the list of time frames
    """
    try:
        # first try in available exchanges
        for exchange_configuration in get_exchange_configurations_from_exchange_name(exchange_name).values():
            return exchange_configuration.exchange_manager.client_time_frames
    except KeyError:
        async with exchanges.temporary_exchange_wrapper(exchange_name, exchange_lib) as wrapper:
            return list(await wrapper.get_available_time_frames())


def get_exchange_available_required_time_frames(exchange_name: str, exchange_id: str) -> list:
    return exchanges.Exchanges.instance().get_exchange(exchange_name, exchange_id).available_required_time_frames


# prefer get_exchange_configurations_from_exchange_name when possible
def get_exchange_configuration_from_exchange_id(exchange_id: str) -> exchanges.ExchangeConfiguration:
    for exchange_configs in exchanges.Exchanges.instance().exchanges.values():
        for exchange_config in exchange_configs.values():
            if exchange_config.exchange_manager.id == exchange_id:
                return exchange_config
    raise KeyError(f"No exchange configuration with id: {exchange_id}")


# prefer get_exchange_manager_from_exchange_name_and_id when possible
def get_exchange_manager_from_exchange_id(exchange_id: str) -> object:
    try:
        return get_exchange_configuration_from_exchange_id(exchange_id).exchange_manager
    except KeyError:
        raise KeyError(f"No exchange manager with id: {exchange_id}")


def get_exchange_managers_from_exchange_ids(exchange_ids) -> list:
    return [get_exchange_manager_from_exchange_id(manager_id) for manager_id in exchange_ids]


def get_trading_exchanges(exchange_managers) -> list:
    return [exchange_manager for exchange_manager in exchange_managers
            if exchange_manager.is_trading]


def is_exchange_trading(exchange_manager) -> bool:
    return exchange_manager.is_trading


def get_exchange_manager_id(exchange_manager) -> str:
    return exchange_manager.id


def get_exchange_manager_is_sandboxed(exchange_manager) -> bool:
    return exchange_manager.is_sandboxed


def get_exchange_current_time(exchange_manager) -> float:
    return exchange_manager.exchange.get_exchange_current_time()


def get_exchange_allowed_time_lag(exchange_manager) -> float:
    return exchange_manager.exchange.allowed_time_lag


def get_exchange_id_from_matrix_id(exchange_name: str, matrix_id: str) -> str:
    for exchange_configuration in get_exchange_configurations_from_exchange_name(exchange_name).values():
        if exchange_configuration.matrix_id == matrix_id:
            return get_exchange_manager_id(exchange_configuration.exchange_manager)
    return None


def get_matrix_id_from_exchange_id(exchange_name: str, exchange_id: str) -> str:
    for exchange_configuration in get_exchange_configurations_from_exchange_name(exchange_name).values():
        if exchange_configuration.id == exchange_id:
            return exchange_configuration.matrix_id
    return None


def get_all_exchange_ids_from_matrix_id(matrix_id) -> list:
    return [
        exchange_configuration.id
        for exchange_configuration in exchanges.Exchanges.instance().get_all_exchanges()
        if exchange_configuration.matrix_id == matrix_id
    ]


def get_exchange_configuration_from_exchange(exchange_name: str, exchange_id: str) -> exchanges.ExchangeConfiguration:
    return exchanges.Exchanges.instance().get_exchange(exchange_name, exchange_id)


def get_all_exchange_ids_with_same_matrix_id(exchange_name: str, exchange_id: str) -> list:
    """
    Used to get all the exchange ids for the same octobot instance represented by the matrix_id
    :param exchange_name: the exchange name
    :param exchange_id: the exchange id
    :return: the exchange ids for the same matrix id
    """
    return get_all_exchange_ids_from_matrix_id(
        get_exchange_configuration_from_exchange(exchange_name, exchange_id).matrix_id)


def get_exchange_names() -> list:
    return exchanges.Exchanges.instance().get_exchange_names()


def get_exchange_ids() -> list:
    return exchanges.Exchanges.instance().get_exchange_ids()


def get_exchange_name(exchange_manager) -> str:
    return exchange_manager.get_exchange_name()


def has_only_ohlcv(exchange_importers):
    return exchanges.ExchangeSimulator.get_real_available_data(exchange_importers) == \
           set(exchange_data.SIMULATOR_PRODUCERS_TO_POSSIBLE_DATA_TYPE[octobot_trading.constants.OHLCV_CHANNEL])


def get_is_backtesting(exchange_manager) -> bool:
    return exchange_manager.is_backtesting


def get_backtesting_data_files(exchange_manager) -> list:
    if not get_is_backtesting(exchange_manager):
        raise RuntimeError("Require a backtesting exchange manager")
    return exchange_manager.exchange.get_backtesting_data_files()


def get_backtesting_data_file(exchange_manager, symbol, time_frame) -> str:
    if not get_is_backtesting(exchange_manager):
        raise RuntimeError("Require a backtesting exchange manager")
    return backtesting_api.get_data_file_from_importers(
        exchange_manager.exchange.connector.exchange_importers, symbol, time_frame
    )


def get_has_websocket(exchange_manager) -> bool:
    return exchange_manager.has_websocket


def supports_websockets(exchange_name: str, tentacles_setup_config) -> bool:
    return exchanges.supports_websocket(exchange_name, tentacles_setup_config)


def get_trading_pairs(exchange_manager) -> list:
    return exchange_manager.exchange_config.traded_symbol_pairs


def get_watched_timeframes(exchange_manager) -> list:
    return exchange_manager.exchange_config.traded_time_frames


def get_base_currency(exchange_manager, pair) -> str:
    return exchange_manager.exchange.get_pair_cryptocurrency(pair)


def get_fees(exchange_manager, symbol) -> dict:
    return exchange_manager.exchange.get_fees(symbol)


def get_max_handled_pair_with_time_frame(exchange_manager) -> int:
    return exchange_manager.exchange.get_max_handled_pair_with_time_frame()


def get_currently_handled_pair_with_time_frame(exchange_manager) -> int:
    return exchange_manager.get_currently_handled_pair_with_time_frame()


def get_required_historical_candles_count(exchange_manager) -> int:
    return exchange_manager.exchange_config.required_historical_candles_count \
        if exchange_manager.exchange_config.required_historical_candles_count > \
        octobot_trading.constants.DEFAULT_CANDLE_HISTORY_SIZE else octobot_trading.constants.DEFAULT_CANDLE_HISTORY_SIZE


def is_overloaded(exchange_manager) -> bool:
    return exchange_manager.get_is_overloaded()


async def is_compatible_account(exchange_name: str, exchange_config: dict, tentacles_setup_config) -> (bool, str):
    return await exchanges.is_compatible_account(exchange_name, exchange_config, tentacles_setup_config)


def is_sponsoring(exchange_name: str) -> bool:
    return trading_backend.is_sponsoring(exchange_name)


def is_valid_account(exchange_manager) -> bool:
    return exchange_manager.is_valid_account


def get_historical_ohlcv(exchange_manager, symbol, time_frame, start_time, end_time):
    return exchanges.get_historical_ohlcv(exchange_manager, symbol, time_frame, start_time, end_time)


def get_bot_id(exchange_manager):
    return exchange_manager.bot_id


def cancel_ccxt_throttle_task():
    for task in asyncio.all_tasks():
        # manually cancel ccxt async throttle task since it apparently can't be cancelled otherwise
        if str(task._coro).startswith("<coroutine object Throttler.looper at"):
            task.cancel()



#  Drakkar-Software OctoBot
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import decimal
import uuid

import pytest

import octobot_trading.enums as enums
import octobot_trading.errors as errors
import octobot_trading.constants as constants
import octobot_trading.personal_data.transactions.types as transaction_types

from tests.exchanges import backtesting_trader, backtesting_config, backtesting_exchange_manager, fake_backtesting
from tests import event_loop

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio

TRANSACTION_SYMBOL = "BTC/USDT"
TRANSACTION_CURRENCY = "BTC"


async def test_initialized(backtesting_trader):
    _, exchange_manager, _ = backtesting_trader
    assert exchange_manager.exchange_personal_data.transactions_manager.is_initialized


async def test_insert_transaction_instance(backtesting_trader):
    _, exchange_manager, _ = backtesting_trader
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 0
    t_id = str(uuid.uuid4())
    t_id_2 = str(uuid.uuid4())
    transaction = transaction_types.RealisedPnlTransaction(
        exchange_name=exchange_manager.exchange_name,
        creation_time=exchange_manager.exchange.get_exchange_current_time(),
        transaction_type=enums.TransactionType.REALISED_PNL,
        currency=TRANSACTION_CURRENCY,
        symbol=TRANSACTION_SYMBOL,
        realised_pnl=constants.ZERO,
        closed_quantity=constants.ONE,
        cumulated_closed_quantity=constants.ONE,
        first_entry_time=constants.ONE,
        average_entry_price=constants.ONE,
        average_exit_price=constants.ONE,
        order_exit_price=constants.ONE,
        leverage=constants.ONE,
        side=enums.PositionSide.SHORT,
        trigger_source=enums.PNLTransactionSource.LIMIT_ORDER)
    transaction.set_transaction_id(t_id)
    transaction_2 = transaction_types.BlockchainTransaction(
        exchange_name=exchange_manager.exchange_name,
        creation_time=exchange_manager.exchange.get_exchange_current_time(),
        currency=TRANSACTION_CURRENCY,
        transaction_type=enums.TransactionType.BLOCKCHAIN_DEPOSIT,
        blockchain_type="Test",
        blockchain_transaction_id=t_id,
        blockchain_transaction_status=enums.BlockchainTransactionStatus.CONFIRMING,
        source_address="0x123456789",
        destination_address=None,
        quantity=decimal.Decimal(50),
        transaction_fee=decimal.Decimal(0.1)
    )
    transaction_2.set_transaction_id(t_id)

    # succeed to upsert unknown transaction
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 0
    exchange_manager.exchange_personal_data.transactions_manager.insert_transaction_instance(transaction)
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 1

    # can't add transaction with the same transaction id without replacing it --> DuplicateTransactionIdError
    with pytest.raises(errors.DuplicateTransactionIdError):
        exchange_manager.exchange_personal_data.transactions_manager.insert_transaction_instance(transaction_2)
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 1
    assert list(exchange_manager.exchange_personal_data.transactions_manager.transactions.values())[-1] is transaction

    # can replace a transaction with the same transaction id
    exchange_manager.exchange_personal_data.transactions_manager.insert_transaction_instance(transaction_2,
                                                                                             replace_if_exists=True)
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 1

    # succeed to upsert a transaction with a new id
    transaction.set_transaction_id(t_id_2)
    exchange_manager.exchange_personal_data.transactions_manager.insert_transaction_instance(transaction)
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 2

    # check transaction instances
    assert list(exchange_manager.exchange_personal_data.transactions_manager.transactions.values())[0] is transaction_2
    assert list(exchange_manager.exchange_personal_data.transactions_manager.transactions.values())[-1] is transaction


async def test_get_transaction(backtesting_trader):
    _, exchange_manager, _ = backtesting_trader
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 0
    transaction = transaction_types.RealisedPnlTransaction(
        exchange_name=exchange_manager.exchange_name,
        creation_time=exchange_manager.exchange.get_exchange_current_time(),
        transaction_type=enums.TransactionType.REALISED_PNL,
        currency=TRANSACTION_CURRENCY,
        symbol=TRANSACTION_SYMBOL,
        side=enums.PositionSide.BOTH,
        realised_pnl=constants.ZERO,
        closed_quantity=constants.ONE,
        cumulated_closed_quantity=constants.ONE,
        first_entry_time=constants.ONE,
        average_entry_price=constants.ONE,
        average_exit_price=constants.ONE,
        order_exit_price=constants.ONE,
        leverage=constants.ONE,
        trigger_source=enums.PNLTransactionSource.LIMIT_ORDER)
    transaction_2 = transaction_types.RealisedPnlTransaction(
        exchange_name=exchange_manager.exchange_name,
        creation_time=exchange_manager.exchange.get_exchange_current_time(),
        transaction_type=enums.TransactionType.REALISED_PNL,
        currency=TRANSACTION_CURRENCY,
        symbol=TRANSACTION_SYMBOL,
        side=enums.PositionSide.LONG,
        realised_pnl=constants.ZERO,
        closed_quantity=constants.ONE,
        cumulated_closed_quantity=constants.ONE,
        first_entry_time=constants.ONE,
        average_entry_price=constants.ONE,
        average_exit_price=constants.ONE,
        order_exit_price=constants.ONE,
        leverage=constants.ONE,
        trigger_source=enums.PNLTransactionSource.LIMIT_ORDER)
    exchange_manager.exchange_personal_data.transactions_manager.insert_transaction_instance(transaction)

    # succeed to return transaction instance
    assert exchange_manager.exchange_personal_data.transactions_manager. \
               get_transaction(transaction.transaction_id) is transaction

    # can't get unknown transaction id
    with pytest.raises(KeyError):
        exchange_manager.exchange_personal_data.transactions_manager.get_transaction(str(uuid.uuid4()))

    # can't get a transaction which has not been added
    with pytest.raises(KeyError):
        exchange_manager.exchange_personal_data.transactions_manager.get_transaction(transaction_2.transaction_id)

    # succeed to return transaction instances
    exchange_manager.exchange_personal_data.transactions_manager.insert_transaction_instance(transaction_2)
    assert exchange_manager.exchange_personal_data.transactions_manager. \
               get_transaction(transaction.transaction_id) is transaction
    assert exchange_manager.exchange_personal_data.transactions_manager. \
               get_transaction(transaction_2.transaction_id) is transaction_2


async def test_update_transaction_id(backtesting_trader):
    _, exchange_manager, _ = backtesting_trader
    t_id = str(uuid.uuid4())
    t_id_2 = str(uuid.uuid4())
    t_id_3 = str(uuid.uuid4())
    t_id_4 = str(uuid.uuid4())
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 0
    transaction = transaction_types.RealisedPnlTransaction(
        exchange_name=exchange_manager.exchange_name,
        creation_time=exchange_manager.exchange.get_exchange_current_time(),
        transaction_type=enums.TransactionType.REALISED_PNL,
        currency=TRANSACTION_CURRENCY,
        symbol=TRANSACTION_SYMBOL,
        side=enums.PositionSide.BOTH,
        realised_pnl=constants.ZERO,
        closed_quantity=constants.ONE,
        cumulated_closed_quantity=constants.ONE,
        first_entry_time=constants.ONE,
        average_entry_price=constants.ONE,
        average_exit_price=constants.ONE,
        order_exit_price=constants.ONE,
        leverage=constants.ONE,
        trigger_source=enums.PNLTransactionSource.LIMIT_ORDER)
    transaction.set_transaction_id(t_id)
    transaction_2 = transaction_types.RealisedPnlTransaction(
        exchange_name=exchange_manager.exchange_name,
        creation_time=exchange_manager.exchange.get_exchange_current_time(),
        transaction_type=enums.TransactionType.REALISED_PNL,
        currency=TRANSACTION_CURRENCY,
        symbol=TRANSACTION_SYMBOL,
        side=enums.PositionSide.SHORT,
        realised_pnl=constants.ZERO,
        closed_quantity=constants.ONE,
        cumulated_closed_quantity=constants.ONE,
        first_entry_time=constants.ONE,
        average_entry_price=constants.ONE,
        average_exit_price=constants.ONE,
        order_exit_price=constants.ONE,
        leverage=constants.ONE,
        trigger_source=enums.PNLTransactionSource.LIMIT_ORDER)
    transaction_2.set_transaction_id(t_id_2)

    # Add transaction instances to TransactionsManager.transactions
    exchange_manager.exchange_personal_data.transactions_manager.insert_transaction_instance(transaction)
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 1
    exchange_manager.exchange_personal_data.transactions_manager.insert_transaction_instance(transaction_2)
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 2

    # Update the first transaction id and see if the previous entry was properly deleted
    exchange_manager.exchange_personal_data.transactions_manager.update_transaction_id(t_id, t_id_3)
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 2
    assert transaction.transaction_id == t_id_3
    assert list(exchange_manager.exchange_personal_data.transactions_manager.transactions.values())[-1] is transaction
    assert list(exchange_manager.exchange_personal_data.transactions_manager.transactions.values())[-1].transaction_id \
           == t_id_3

    # ensure that the first transaction_id doesn't exist anymore
    with pytest.raises(KeyError):
        exchange_manager.exchange_personal_data.transactions_manager.get_transaction(t_id)

    # try to update with an already used transaction id without replacing it
    with pytest.raises(errors.DuplicateTransactionIdError):
        exchange_manager.exchange_personal_data.transactions_manager.update_transaction_id(t_id_2, t_id_3)

    # check if the transaction keeps its old transaction id
    assert transaction_2.transaction_id == t_id_2

    # check if transaction_2 was not dropped from the list
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 2
    exchange_manager.exchange_personal_data.transactions_manager.get_transaction(t_id_2)

    # try to update with an already used transaction id with replacement
    exchange_manager.exchange_personal_data.transactions_manager.update_transaction_id(t_id_2, t_id_3,
                                                                                       replace_if_exists=True)
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 1  # transaction replaced by transaction_2
    assert transaction_2.transaction_id == t_id_3
    assert list(exchange_manager.exchange_personal_data.transactions_manager.transactions.values())[-1] is transaction_2
    assert list(exchange_manager.exchange_personal_data.transactions_manager.transactions.values())[
               -1].transaction_id == t_id_3

    # try to update an unknown transaction
    with pytest.raises(KeyError):
        exchange_manager.exchange_personal_data.transactions_manager.update_transaction_id(str(uuid.uuid4()),
                                                                                           str(uuid.uuid4()))
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 1

    # restore transaction in list
    transaction.set_transaction_id(t_id_4)
    exchange_manager.exchange_personal_data.transactions_manager.insert_transaction_instance(transaction)
    assert len(exchange_manager.exchange_personal_data.transactions_manager.transactions) == 2



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.constants as constants
import octobot_commons.logging as logging

import octobot_trading.errors as errors
import octobot_trading.modes as modes

LOGGER_TAG = "TradingModeFactory"


async def create_trading_modes(config: dict,
                               exchange_manager: object,
                               trading_mode_class: modes.AbstractTradingMode.__class__,
                               bot_id: str) -> list:
    is_symbol_wildcard = trading_mode_class.get_is_symbol_wildcard()
    if is_symbol_wildcard or (not is_symbol_wildcard and exchange_manager.exchange_config.traded_symbol_pairs):
        return await _create_trading_modes(trading_mode_class=trading_mode_class,
                                           config=config,
                                           exchange_manager=exchange_manager,
                                           cryptocurrencies=exchange_manager.exchange_config.traded_cryptocurrencies,
                                           symbols=exchange_manager.exchange_config.traded_symbol_pairs,
                                           time_frames=exchange_manager.exchange_config.traded_time_frames,
                                           bot_id=bot_id)
    # Do not create no symbol wildcard trading mode if no trading pair is available
    raise errors.TradingModeIncompatibility(
        f"As non symbol-wildcard trading mode, {trading_mode_class.get_name()} requires "
        f"at least one exchange trading pair to be initialized. "
        f"None of the required pairs are available on {exchange_manager.exchange_name}.")


async def _create_trading_modes(trading_mode_class: modes.AbstractTradingMode.__class__,
                                config: dict,
                                exchange_manager: object,
                                cryptocurrencies: dict = None,
                                symbols: list = None,
                                time_frames: list = None,
                                bot_id: str = None) -> list:
    return [
        await create_trading_mode(trading_mode_class=trading_mode_class,
                                  config=config,
                                  exchange_manager=exchange_manager,
                                  cryptocurrency=cryptocurrency,
                                  symbol=symbol,
                                  time_frame=time_frame,
                                  bot_id=bot_id)
        for cryptocurrency in _get_cryptocurrencies_to_create(trading_mode_class, cryptocurrencies)
        for symbol in _get_symbols_to_create(trading_mode_class, cryptocurrencies, cryptocurrency, symbols)
        for time_frame in _get_time_frames_to_create(trading_mode_class, time_frames)
    ]


async def create_trading_mode(trading_mode_class: modes.AbstractTradingMode.__class__,
                              config: dict,
                              exchange_manager: object,
                              cryptocurrency: str = None,
                              symbol: str = None,
                              time_frame: object = None,
                              bot_id: str = None) -> modes.AbstractTradingMode:
    try:
        trading_mode: modes.AbstractTradingMode = trading_mode_class(config, exchange_manager)
        trading_mode.cryptocurrency = cryptocurrency
        trading_mode.symbol = symbol
        trading_mode.time_frame = time_frame
        trading_mode.bot_id = bot_id
        await trading_mode.initialize()
        logging.get_logger(f"{LOGGER_TAG}[{exchange_manager.exchange_name}]") \
            .debug(f"{trading_mode.get_name()} started for "
                   f"[cryptocurrency={cryptocurrency if cryptocurrency else constants.CONFIG_WILDCARD},"
                   f" symbol={symbol if symbol else constants.CONFIG_WILDCARD},"
                   f" time_frame={time_frame if time_frame else constants.CONFIG_WILDCARD}]")
        return trading_mode
    except RuntimeError as e:
        logging.get_logger(LOGGER_TAG).error(e.args[0])
        raise e


def _get_cryptocurrencies_to_create(trading_mode_class, cryptocurrencies):
    return list(cryptocurrencies.keys()) \
        if cryptocurrencies and not trading_mode_class.get_is_cryptocurrency_wildcard() else [None]


def _get_symbols_to_create(trading_mode_class, cryptocurrencies, cryptocurrency, symbols):
    currency_symbols = symbols
    if cryptocurrency is not None:
        currency_symbols = cryptocurrencies.get(cryptocurrency, [])
    return currency_symbols if currency_symbols and not trading_mode_class.get_is_symbol_wildcard() else [None]


def _get_time_frames_to_create(trading_mode_class, time_frames):
    return time_frames if time_frames and not trading_mode_class.get_is_time_frame_wildcard() else [None]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import uuid

import octobot_trading.constants
import octobot_trading.exchanges.traders.trader as trader
import octobot_trading.util as util


class TraderSimulator(trader.Trader):
    """
    TraderSimulator has a role of exchange response simulator
    - During order creation / filling / canceling process
    """

    NO_HISTORY_MESSAGE = "Starting a fresh new trading simulation session using trader simulator initial portfolio " \
                         "in configuration."

    def __init__(self, config, exchange_manager):
        self.simulate = True
        super().__init__(config, exchange_manager)

        self.trader_type_str = octobot_trading.constants.SIMULATOR_TRADER_STR

    @staticmethod
    def enabled(config):
        return util.is_trader_simulator_enabled(config)

    def parse_order_id(self, order_id):
        return str(uuid.uuid4()) if order_id is None else order_id



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
from ccxt import Exchange

from octobot_commons.constants import MINUTE_TO_SECONDS, MSECONDS_TO_SECONDS
from octobot_commons.enums import TimeFrames, TimeFramesMinutes
from octobot_trading.enums import ExchangeConstantsTickersColumns as Ectc
from tests_additional.real_exchanges import get_exchange_manager


class RealExchangeTester:
    # enter exchange name as a class variable here
    EXCHANGE_NAME = None
    SYMBOL = None
    SYMBOL_2 = None
    # default is 1h, change if necessary
    TIME_FRAME = TimeFrames.ONE_HOUR

    # Public methods: to be implemented as tests
    # Use await self._[method_name] to get the test request result
    # ex: market_status = await self.get_market_status()

    # unauthenticated API
    async def test_time_frames(self):
        pass

    async def test_get_market_status(self):
        pass

    async def test_get_symbol_prices(self):
        pass

    async def test_get_kline_price(self):
        pass

    async def test_get_order_book(self):
        pass

    async def test_get_recent_trades(self):
        pass

    async def test_get_price_ticker(self):
        pass

    async def test_get_all_currencies_price_ticker(self):
        pass

    # authenticated API
    # TODO
    # async def test_get_balance(self):
    #     pass
    #
    # async def test_get_order(self):
    #     pass
    #
    # async def test_get_all_orders(self):
    #     pass
    #
    # async def test_get_open_orders(self):
    #     pass
    #
    # async def test_get_closed_orders(self):
    #     pass
    #
    # async def test_get_my_recent_trades(self):
    #     pass
    #
    # async def test_cancel_order(self):
    #     pass
    #
    # async def test_create_order(self):
    #     pass

    async def time_frames(self):
        async with get_exchange_manager(self.EXCHANGE_NAME) as exchange_manager:
            return exchange_manager.exchange.time_frames

    async def get_market_statuses(self):
        # return 2 different market status with different traded pairs to reduce possible
        # side effects using only one pair.
        async with get_exchange_manager(self.EXCHANGE_NAME) as exchange_manager:
            return exchange_manager.exchange.get_market_status(self.SYMBOL), \
                   exchange_manager.exchange.get_market_status(self.SYMBOL_2)

    async def get_symbol_prices(self, limit=None, **kwargs):
        async with get_exchange_manager(self.EXCHANGE_NAME) as exchange_manager:
            return await exchange_manager.exchange.get_symbol_prices(self.SYMBOL, self.TIME_FRAME,
                                                                     limit=limit, **kwargs)

    async def get_kline_price(self, **kwargs):
        async with get_exchange_manager(self.EXCHANGE_NAME) as exchange_manager:
            return await exchange_manager.exchange.get_kline_price(self.SYMBOL, self.TIME_FRAME, **kwargs)

    async def get_order_book(self, **kwargs):
        async with get_exchange_manager(self.EXCHANGE_NAME) as exchange_manager:
            return await exchange_manager.exchange.get_order_book(self.SYMBOL, **kwargs)

    async def get_recent_trades(self, limit=50):
        async with get_exchange_manager(self.EXCHANGE_NAME) as exchange_manager:
            return await exchange_manager.exchange.get_recent_trades(self.SYMBOL, limit=limit)

    async def get_price_ticker(self):
        async with get_exchange_manager(self.EXCHANGE_NAME) as exchange_manager:
            return await exchange_manager.exchange.get_price_ticker(self.SYMBOL)

    async def get_all_currencies_price_ticker(self, **kwargs):
        async with get_exchange_manager(self.EXCHANGE_NAME) as exchange_manager:
            return await exchange_manager.exchange.get_all_currencies_price_ticker(**kwargs)

    def get_allowed_time_delta(self):
        return TimeFramesMinutes[self.TIME_FRAME] * MINUTE_TO_SECONDS * MSECONDS_TO_SECONDS * 1.3

    @staticmethod
    def get_time():
        return Exchange.milliseconds()

    @staticmethod
    def ensure_elements_order(elements, sort_key, reverse=False):
        assert sorted(elements, key=lambda x: x[sort_key], reverse=reverse) == elements

    @staticmethod
    def check_ticker_typing(ticker, check_open=True, check_high=True, check_low=True,
                            check_close=True, check_base_volume=True, check_timestamp=True):
        if check_open:
            assert isinstance(ticker[Ectc.OPEN.value], (float, int))
        if check_high:
            assert isinstance(ticker[Ectc.HIGH.value], (float, int))
        if check_low:
            assert isinstance(ticker[Ectc.LOW.value], (float, int))
        if check_close:
            assert isinstance(ticker[Ectc.CLOSE.value], (float, int))
        if check_base_volume:
            assert isinstance(ticker[Ectc.BASE_VOLUME.value], (float, int))
        if check_timestamp:
            assert isinstance(ticker[Ectc.TIMESTAMP.value], (float, int))



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.exchanges.connectors import exchange_simulator
from octobot_trading.exchanges.connectors.exchange_simulator import (
    ExchangeSimulator,
)
from octobot_trading.exchanges.connectors import ccxt_exchange
from octobot_trading.exchanges.connectors.ccxt_exchange import (
    CCXTExchange,
)
from octobot_trading.exchanges.connectors import abstract_websocket_connector
from octobot_trading.exchanges.connectors.abstract_websocket_connector import (
    AbstractWebsocketConnector,
)
from octobot_trading.exchanges.connectors import ccxt_websocket_connector
from octobot_trading.exchanges.connectors.ccxt_websocket_connector import (
    CCXTWebsocketConnector,
)
from octobot_trading.exchanges.connectors import cryptofeed_websocket_connector
from octobot_trading.exchanges.connectors.cryptofeed_websocket_connector import (
    CryptofeedWebsocketConnector,
)

__all__ = [
    "ExchangeSimulator",
    "CCXTExchange",
    "AbstractWebsocketConnector",
    "CCXTWebsocketConnector",
    "CryptofeedWebsocketConnector",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.personal_data.trades cimport channel
from octobot_trading.personal_data.trades.channel cimport (
    TradesProducer,
    TradesChannel,
    TradesUpdater,
)
from octobot_trading.personal_data.trades cimport trade
from octobot_trading.personal_data.trades.trade cimport (
    Trade,
)
from octobot_trading.personal_data.trades cimport trades_manager
from octobot_trading.personal_data.trades.trades_manager cimport (
    TradesManager,
)
from octobot_trading.personal_data.trades cimport trade_factory
from octobot_trading.personal_data.trades.trade_factory cimport (
    create_trade_instance_from_raw,
    create_trade_from_order,
    create_trade_instance,
)

__all__ = [
    "TradesManager",
    "TradesProducer",
    "TradesChannel",
    "create_trade_instance_from_raw",
    "create_trade_from_order",
    "create_trade_instance",
    "TradesUpdater",
    "Trade",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

import octobot_trading.enums

from octobot_trading.personal_data import state
from octobot_trading.personal_data.state import (
    State,
)
from octobot_trading.personal_data import orders
from octobot_trading.personal_data.orders import (
    Order,
    parse_order_type,
    is_valid,
    get_min_max_amounts,
    check_cost,
    get_futures_max_order_size,
    get_max_order_quantity_for_price,
    total_fees_from_order_dict,
    get_fees_for_currency,
    parse_raw_fees,
    parse_order_status,
    parse_is_cancelled,
    get_pnl_transaction_source_from_order,
    is_stop_order,
    create_as_chained_order,
    ensure_orders_relevancy,
    get_order_quantity_currency,
    is_associated_pending_order,
    apply_pending_order_from_created_order,
    get_up_to_date_price,
    get_pre_order_data,
    OrderState,
    OrderGroup,
    BalancedTakeProfitAndStopOrderGroup,
    OneCancelsTheOtherOrderGroup,
    OrdersUpdater,
    adapt_price,
    decimal_adapt_price,
    adapt_quantity,
    decimal_adapt_quantity,
    trunc_with_n_decimal_digits,
    decimal_trunc_with_n_decimal_digits,
    adapt_order_quantity_because_quantity,
    decimal_adapt_order_quantity_because_quantity,
    adapt_order_quantity_because_price,
    decimal_adapt_order_quantity_because_price,
    split_orders,
    decimal_split_orders,
    check_and_adapt_order_details_if_necessary,
    decimal_check_and_adapt_order_details_if_necessary,
    add_dusts_to_quantity_if_necessary,
    decimal_add_dusts_to_quantity_if_necessary,
    create_order_from_raw,
    create_order_instance_from_raw,
    create_order_from_type,
    create_order_instance,
    OrdersProducer,
    OrdersChannel,
    OrdersManager,
    OrdersUpdaterSimulator,
    CloseOrderState,
    CancelOrderState,
    OpenOrderState,
    create_order_state,
    FillOrderState,
    PendingCreationOrderState,
    UnknownOrder,
    MarketOrder,
    SellMarketOrder,
    BuyMarketOrder,
    BuyLimitOrder,
    SellLimitOrder,
    LimitOrder,
    TakeProfitOrder,
    StopLossOrder,
    StopLossLimitOrder,
    TakeProfitLimitOrder,
    TrailingStopOrder,
    TrailingStopLimitOrder,
)
from octobot_trading.personal_data import portfolios
from octobot_trading.personal_data.portfolios import (
    BalanceUpdaterSimulator,
    BalanceProfitabilityUpdaterSimulator,
    create_portfolio_from_exchange_manager,
    BalanceUpdater,
    BalanceProfitabilityUpdater,
    PortfolioProfitability,
    Portfolio,
    BalanceProducer,
    BalanceChannel,
    BalanceProfitabilityProducer,
    BalanceProfitabilityChannel,
    SubPortfolio,
    PortfolioManager,
    PortfolioValueHolder,
    FuturePortfolio,
    MarginPortfolio,
    SpotPortfolio,
    Asset,
    FutureAsset,
    MarginAsset,
    SpotAsset,
    parse_decimal_portfolio,
    parse_decimal_config_portfolio,
    portfolio_to_float,
    create_historical_asset_value_from_dict_like_object,
    get_draw_down,
    get_coefficient_of_determination,
    HistoricalAssetValue,
    HistoricalPortfolioValueManager,
)
from octobot_trading.personal_data import positions
from octobot_trading.personal_data.positions import (
    PositionState,
    PositionsProducer,
    PositionsChannel,
    PositionsUpdaterSimulator,
    Position,
    LinearPosition,
    InversePosition,
    PositionsUpdater,
    PositionsManager,
    create_position_instance_from_raw,
    create_position_from_type,
    create_symbol_position,
    parse_position_status,
    LiquidatePositionState,
    OpenPositionState,
    create_position_state,
)
from octobot_trading.personal_data import trades
from octobot_trading.personal_data.trades import (
    TradesManager,
    TradesProducer,
    TradesChannel,
    create_trade_instance_from_raw,
    create_trade_from_order,
    create_trade_instance,
    TradesUpdater,
    Trade,
    compute_win_rate,
)
from octobot_trading.personal_data import transactions
from octobot_trading.personal_data.transactions import (
    TransactionsManager,
    Transaction,
    BlockchainTransaction,
    FeeTransaction,
    RealisedPnlTransaction,
    TransferTransaction,
    create_blockchain_transaction,
    create_realised_pnl_transaction,
    create_fee_transaction,
    create_transfer_transaction,
)
from octobot_trading.personal_data import exchange_personal_data
from octobot_trading.personal_data.exchange_personal_data import (
    ExchangePersonalData,
)

AUTHENTICATED_UPDATER_PRODUCERS = [
    BalanceUpdater,
    OrdersUpdater,
    TradesUpdater,
    PositionsUpdater,
    BalanceProfitabilityUpdater
]
AUTHENTICATED_UPDATER_SIMULATOR_PRODUCERS = [
    OrdersUpdaterSimulator,
    BalanceProfitabilityUpdater,
    PositionsUpdaterSimulator
]

TraderOrderTypeClasses = {
    octobot_trading.enums.TraderOrderType.BUY_MARKET: BuyMarketOrder,
    octobot_trading.enums.TraderOrderType.BUY_LIMIT: BuyLimitOrder,
    octobot_trading.enums.TraderOrderType.TAKE_PROFIT: TakeProfitOrder,
    octobot_trading.enums.TraderOrderType.TAKE_PROFIT_LIMIT: TakeProfitLimitOrder,
    octobot_trading.enums.TraderOrderType.TRAILING_STOP: TrailingStopOrder,
    octobot_trading.enums.TraderOrderType.TRAILING_STOP_LIMIT: TrailingStopLimitOrder,
    octobot_trading.enums.TraderOrderType.STOP_LOSS: StopLossOrder,
    octobot_trading.enums.TraderOrderType.STOP_LOSS_LIMIT: StopLossLimitOrder,
    octobot_trading.enums.TraderOrderType.SELL_MARKET: SellMarketOrder,
    octobot_trading.enums.TraderOrderType.SELL_LIMIT: SellLimitOrder,
    octobot_trading.enums.TraderOrderType.UNKNOWN: UnknownOrder,
}

TraderPositionTypeClasses = {
    octobot_trading.enums.FutureContractType.LINEAR_PERPETUAL: LinearPosition,
    octobot_trading.enums.FutureContractType.INVERSE_PERPETUAL: InversePosition,
    octobot_trading.enums.FutureContractType.LINEAR_EXPIRABLE: LinearPosition,
    octobot_trading.enums.FutureContractType.INVERSE_EXPIRABLE: InversePosition,
}

__all__ = [
    "State",
    "Order",
    "parse_order_type",
    "is_valid",
    "get_min_max_amounts",
    "check_cost",
    "get_futures_max_order_size",
    "get_max_order_quantity_for_price",
    "total_fees_from_order_dict",
    "get_fees_for_currency",
    "parse_raw_fees",
    "parse_order_status",
    "parse_is_cancelled",
    "get_pnl_transaction_source_from_order",
    "is_stop_order",
    "create_as_chained_order",
    "ensure_orders_relevancy",
    "get_order_quantity_currency",
    "is_associated_pending_order",
    "apply_pending_order_from_created_order",
    "get_up_to_date_price",
    "get_pre_order_data",
    "OrderState",
    "OrderGroup",
    "BalancedTakeProfitAndStopOrderGroup",
    "OneCancelsTheOtherOrderGroup",
    "OrdersUpdater",
    "adapt_price",
    "decimal_adapt_price",
    "adapt_quantity",
    "decimal_adapt_quantity",
    "trunc_with_n_decimal_digits",
    "decimal_trunc_with_n_decimal_digits",
    "adapt_order_quantity_because_quantity",
    "decimal_adapt_order_quantity_because_quantity",
    "adapt_order_quantity_because_price",
    "decimal_adapt_order_quantity_because_price",
    "split_orders",
    "decimal_split_orders",
    "check_and_adapt_order_details_if_necessary",
    "decimal_check_and_adapt_order_details_if_necessary",
    "add_dusts_to_quantity_if_necessary",
    "decimal_add_dusts_to_quantity_if_necessary",
    "create_order_from_raw",
    "create_order_instance_from_raw",
    "create_order_from_type",
    "create_order_instance",
    "OrdersProducer",
    "OrdersChannel",
    "OrdersManager",
    "OrdersUpdaterSimulator",
    "CloseOrderState",
    "CancelOrderState",
    "OpenOrderState",
    "create_order_state",
    "FillOrderState",
    "PendingCreationOrderState",
    "UnknownOrder",
    "MarketOrder",
    "SellMarketOrder",
    "BuyMarketOrder",
    "BuyLimitOrder",
    "SellLimitOrder",
    "LimitOrder",
    "TakeProfitOrder",
    "StopLossOrder",
    "StopLossLimitOrder",
    "TakeProfitLimitOrder",
    "TrailingStopOrder",
    "TrailingStopLimitOrder",
    "PositionsProducer",
    "PositionsChannel",
    "BalanceUpdaterSimulator",
    "BalanceProfitabilityUpdaterSimulator",
    "create_portfolio_from_exchange_manager",
    "BalanceUpdater",
    "BalanceProfitabilityUpdater",
    "PortfolioProfitability",
    "Portfolio",
    "BalanceProducer",
    "BalanceChannel",
    "BalanceProfitabilityProducer",
    "BalanceProfitabilityChannel",
    "SubPortfolio",
    "PortfolioManager",
    "PortfolioValueHolder",
    "FuturePortfolio",
    "MarginPortfolio",
    "SpotPortfolio",
    "Asset",
    "FutureAsset",
    "MarginAsset",
    "SpotAsset",
    "parse_decimal_portfolio",
    "parse_decimal_config_portfolio",
    "portfolio_to_float",
    "create_historical_asset_value_from_dict_like_object",
    "get_draw_down",
    "create_historical_asset_value_from_dict_like_object",
    "HistoricalAssetValue",
    "HistoricalPortfolioValueManager",
    "PositionsUpdaterSimulator",
    "Position",
    "LinearPosition",
    "InversePosition",
    "PositionState",
    "LiquidatePositionState",
    "OpenPositionState",
    "create_position_state",
    "PositionsUpdater",
    "PositionsManager",
    "create_position_instance_from_raw",
    "create_position_from_type",
    "create_symbol_position",
    "parse_position_status",
    "TradesManager",
    "TradesProducer",
    "TradesChannel",
    "create_trade_instance_from_raw",
    "create_trade_from_order",
    "create_trade_instance",
    "TradesUpdater",
    "Trade",
    "compute_win_rate",
    "ExchangePersonalData",
    "AUTHENTICATED_UPDATER_PRODUCERS",
    "AUTHENTICATED_UPDATER_SIMULATOR_PRODUCERS",
    "TraderOrderTypeClasses",
    "TraderPositionTypeClasses",
    "TransactionsManager",
    "Transaction",
    "BlockchainTransaction",
    "FeeTransaction",
    "RealisedPnlTransaction",
    "TransferTransaction",
    "create_blockchain_transaction",
    "create_realised_pnl_transaction",
    "create_fee_transaction",
    "create_transfer_transaction",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.personal_data.transactions cimport transaction
from octobot_trading.personal_data.transactions.transaction cimport (
    Transaction,
)
from octobot_trading.personal_data.transactions cimport transactions_manager
from octobot_trading.personal_data.transactions.transactions_manager cimport (
    TransactionsManager,
)
from octobot_trading.personal_data.transactions cimport transaction_factory
from octobot_trading.personal_data.transactions.transaction_factory cimport (
    create_blockchain_transaction,
    create_realised_pnl_transaction,
    create_fee_transaction,
    create_transfer_transaction,
)
from octobot_trading.personal_data.transactions cimport types
from octobot_trading.personal_data.transactions.types cimport (
    BlockchainTransaction,
    FeeTransaction,
    RealisedPnlTransaction,
    TransferTransaction,
)

__all__ = [
    "TransactionsManager",
    "Transaction",
    "BlockchainTransaction",
    "FeeTransaction",
    "RealisedPnlTransaction",
    "TransferTransaction",
    "create_blockchain_transaction",
    "create_realised_pnl_transaction",
    "create_fee_transaction",
    "create_transfer_transaction",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import async_channel.channels as channels

import octobot_commons.errors as errors
import octobot_commons.channels_name as channels_name

import octobot_trading.exchange_data.kline.channel.kline_updater as kline_updater
import octobot_trading.util as util


class KlineUpdaterSimulator(kline_updater.KlineUpdater):
    def __init__(self, channel, importer):
        super().__init__(channel)
        self.exchange_data_importer = importer
        self.exchange_name = self.channel.exchange_manager.exchange_name

        self.last_timestamp_pushed = 0
        self.time_consumer = None

    async def start(self):
        await self.resume()

    async def handle_timestamp(self, timestamp, **kwargs):
        try:
            for time_frame in self.channel.exchange_manager.exchange_config.traded_time_frames:
                for pair in self.channel.exchange_manager.exchange_config.traded_symbol_pairs:
                    kline_data = await self.exchange_data_importer.get_kline_from_timestamps(
                        exchange_name=self.exchange_name,
                        symbol=pair,
                        time_frame=time_frame,
                        inferior_timestamp=timestamp,
                        limit=1)
                    if kline_data and kline_data[0][0] > self.last_timestamp_pushed:
                        self.last_timestamp_pushed = kline_data[0][0]
                        await self.push(time_frame, pair, kline_data[0][-1])
        except errors.DatabaseNotFoundError as e:
            self.logger.warning(f"Not enough data : {e}")
            await self.pause()
            await self.stop()
        except IndexError as e:
            self.logger.warning(f"Failed to access kline_data : {e}")

    async def pause(self):
        if self.time_consumer is not None:
            await channels.get_chan(channels_name.OctoBotBacktestingChannelsName.TIME_CHANNEL.value) \
                .remove_consumer(self.time_consumer)

    async def stop(self):
        await util.stop_and_pause(self)

    async def resume(self):
        if self.time_consumer is None and not self.channel.is_paused:
            self.time_consumer = await channels.get_chan(channels_name.OctoBotBacktestingChannelsName.
                                                         TIME_CHANNEL.value).new_consumer(self.handle_timestamp)



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import abc

import octobot_commons.constants as common_constants
import octobot_commons.logging as logging
import octobot_commons.tentacles_management as abstract_tentacle

import async_channel.constants as channel_constants

import octobot_tentacles_manager.api as tentacles_manager_api
import octobot_tentacles_manager.configuration as tm_configuration

import octobot_trading.constants as constants
import octobot_trading.enums as enums
import octobot_trading.exchange_channel as exchanges_channel


class AbstractTradingMode(abstract_tentacle.AbstractTentacle):
    __metaclass__ = abc.ABCMeta

    MODE_PRODUCER_CLASSES = []
    MODE_CONSUMER_CLASSES = []

    def __init__(self, config, exchange_manager):
        super().__init__()
        self.logger = logging.get_logger(self.get_name())

        # Global OctoBot configuration
        self.config: dict = config

        # Mode related exchange manager instance
        self.exchange_manager = exchange_manager

        # The id of the OctoBot using this trading mode
        self.bot_id: str = None

        # Evaluator specific config (Is loaded from tentacle specific file)
        self.trading_config: dict = None

        # If this mode is enabled
        self.enabled: bool = True

        # Specified Cryptocurrency for this instance (Should be None if wildcard)
        self.cryptocurrency: str = None

        # Symbol is the cryptocurrency pair (Should be None if wildcard)
        self.symbol: str = None

        # Time_frame is the chart time frame (Should be None if wildcard)
        self.time_frame = None

        # producers is the list of producers created by this trading mode
        self.producers = []

        # producers is the list of consumers created by this trading mode
        self.consumers = []

        # True when this trading mode is waken up only after full candles close
        self.is_triggered_after_candle_close = False

    # Used to know the current state of the trading mode.
    # Overwrite in subclasses
    def get_current_state(self) -> tuple:
        """
        :return: (str, float): (current state description, current state value)
        """
        return "N/A", 0

    @classmethod
    def get_is_cryptocurrency_wildcard(cls) -> bool:
        """
        :return: True if the mode is not cryptocurrency dependant else False
        """
        return True

    @classmethod
    def get_is_symbol_wildcard(cls) -> bool:
        """
        :return: True if the mode is not symbol dependant else False
        """
        return True

    @classmethod
    def get_is_time_frame_wildcard(cls) -> bool:
        """
        :return: True if the mode is not time_frame dependant else False
        """
        return True

    @classmethod
    def get_supported_exchange_types(cls) -> list:
        """
        :return: The list of supported exchange types
        """
        return [
            enums.ExchangeTypes.SPOT
        ]

    def is_trading_signal_emitter(self) -> bool:
        """
        :return: True if the mode should be emitting trading signals
        """
        try:
            return self.trading_config[common_constants.CONFIG_EMIT_TRADING_SIGNALS]
        except KeyError:
            return False

    @classmethod
    def get_is_trading_on_exchange(cls, exchange_name,
                                   tentacles_setup_config: tm_configuration.TentaclesSetupConfiguration) -> bool:
        """
        :return: When returning false, the associated exchange_manager.is_trading will be set to false, which will
        prevent the initialization of trade related elements. Default is True
        """
        return True

    @classmethod
    def get_parent_trading_mode_classes(cls, higher_parent_class_limit=None) -> list:
        return [
            class_type
            for class_type in cls.mro()
            if (higher_parent_class_limit if higher_parent_class_limit else AbstractTradingMode) in class_type.mro()
        ]

    @staticmethod
    def is_backtestable() -> bool:
        """
        Should be overwritten
        :return: True if the TradingMode can be used in a backtesting else False
        """
        return True

    async def initialize(self) -> None:
        """
        Triggers producers and consumers creation
        """
        self.producers = await self.create_producers()
        self.consumers = await self.create_consumers()

    async def stop(self) -> None:
        """
        Stops all producers and consumers
        """
        for producer in self.producers:
            await producer.stop()
        for consumer in self.consumers:
            await consumer.stop()
        self.exchange_manager = None

    async def create_producers(self) -> list:
        """
        Creates the instance of producers listed in MODE_PRODUCER_CLASSES
        :return: the list of producers created
        """
        return [
            await self._create_mode_producer(mode_producer_class)
            for mode_producer_class in self.MODE_PRODUCER_CLASSES
        ]

    async def _create_mode_producer(self, mode_producer_class):
        """
        Creates a new :mode_producer_class: instance and starts it
        :param mode_producer_class: the trading mode producer class to create
        :return: the producer class created
        """
        mode_producer = mode_producer_class(
            exchanges_channel.get_chan(constants.MODE_CHANNEL, self.exchange_manager.id),
            self.config, self, self.exchange_manager)
        await mode_producer.run()
        return mode_producer

    async def create_consumers(self) -> list:
        """
        Creates the instance of consumers listed in MODE_CONSUMER_CLASSES
        :return: the list of consumers created
        """
        return [
            await self._create_mode_consumer(mode_consumer_class)
            for mode_consumer_class in self.MODE_CONSUMER_CLASSES
        ]

    async def _create_mode_consumer(self, mode_consumer_class):
        """
        Creates a new :mode_consumer_class: instance and subscribe this new consumer to the trading mode channel
        :param mode_consumer_class: the trading mode consumer class to create
        :return: the consumer class created
        """
        mode_consumer = mode_consumer_class(self)
        await exchanges_channel.get_chan(constants.MODE_CHANNEL, self.exchange_manager.id).new_consumer(
            consumer_instance=mode_consumer,
            trading_mode_name=self.get_name(),
            cryptocurrency=self.cryptocurrency if self.cryptocurrency else channel_constants.CHANNEL_WILDCARD,
            symbol=self.symbol if self.symbol else channel_constants.CHANNEL_WILDCARD,
            time_frame=self.time_frame if self.time_frame else channel_constants.CHANNEL_WILDCARD)
        return mode_consumer

    def load_config(self) -> None:
        """
        Try to load TradingMode tentacle config.
        Calls set_default_config() if the tentacle config is empty
        """
        # try with this class name
        self.trading_config = tentacles_manager_api.get_tentacle_config(self.exchange_manager.tentacles_setup_config,
                                                                        self.__class__)

        # set default config if nothing found
        if not self.trading_config:
            self.set_default_config()

    # to implement in subclasses if config is necessary
    def set_default_config(self) -> None:
        pass

    """
    Strategy related methods
    """

    @classmethod
    def get_required_strategies_names_and_count(cls,
                                                tentacles_config: tm_configuration.TentaclesSetupConfiguration,
                                                trading_mode_config=None):
        config = trading_mode_config or tentacles_manager_api.get_tentacle_config(tentacles_config, cls)
        if constants.TRADING_MODE_REQUIRED_STRATEGIES in config:
            return config[constants.TRADING_MODE_REQUIRED_STRATEGIES], cls.get_required_strategies_count(config)
        raise Exception(f"'{constants.TRADING_MODE_REQUIRED_STRATEGIES}' is missing in configuration file")

    @classmethod
    def get_default_strategies(cls,
                               tentacles_config: tm_configuration.TentaclesSetupConfiguration,
                               trading_mode_config=None):
        config = trading_mode_config or tentacles_manager_api.get_tentacle_config(tentacles_config, cls)
        if common_constants.TENTACLE_DEFAULT_CONFIG in config:
            return config[common_constants.TENTACLE_DEFAULT_CONFIG]

        strategies_classes, _ = cls.get_required_strategies_names_and_count(tentacles_config, config)
        return strategies_classes

    @classmethod
    def get_required_strategies_count(cls, config):
        min_strategies_count = 1
        if constants.TRADING_MODE_REQUIRED_STRATEGIES_MIN_COUNT in config:
            min_strategies_count = config[constants.TRADING_MODE_REQUIRED_STRATEGIES_MIN_COUNT]
        return min_strategies_count

    @classmethod
    def get_required_candles_count(cls, tentacles_setup_config: tm_configuration.TentaclesSetupConfiguration):
        return tentacles_manager_api.get_tentacle_config(tentacles_setup_config, cls).get(
            constants.CONFIG_CANDLES_HISTORY_SIZE_KEY,
            common_constants.DEFAULT_IGNORED_VALUE
        )
